"""Generate layouts for components in a Pixel dataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import itertools
import os
import tempfile
from pathlib import Path
from typing import Iterable

import pandas as pd
import polars as pl
import pyarrow as pa

from pixelator.common.graph.backends.protocol import SupportedLayoutAlgorithm
from pixelator.pna import read
from pixelator.pna.analysis_engine import (
    PerComponentTask,
)
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset import PNAPixelDataset
from pixelator.pna.pixeldataset.io import PixelFileWriter, PxlFile


class CreateLayout(PerComponentTask):
    """Run one or more layout algorithms on a component."""

    TASK_NAME = "layout"

    def __init__(
        self,
        layout_algorithms: list[SupportedLayoutAlgorithm],
        algorithm_kwargs: dict | None = None,
    ) -> None:
        """Create a new CreateLayout instance.

        :param layout_algorithms: The layout algorithms to run.
        :param algorithm_kwargs: Additional keyword arguments to pass to the layout algorithms.
        """
        super().__init__()
        self._layout_algorithms = layout_algorithms
        self._algorithm_kwargs = algorithm_kwargs or {}
        self.pxl_dataset: PNAPixelDataset | None = None

    def set_dataset(self, pxl_file_path: Path):
        """Specify a dataset to enable analysis being run directly from component IDs."""
        self.pxl_dataset = read(pxl_file_path)

    def run_from_component_id(self, component_id: str):
        """Run the layout on a component specified by its ID.

        This only works when the pxl_dataset is set so that components
        are directly accessible through their IDs.

        :param component_id: The id of the component.
        :return: a LazyFrame containing the layout data.
        """
        edgelist = (
            self.pxl_dataset.filter(components=[component_id])  # type: ignore
            .edgelist()
            .to_record_batches()
        )
        res = self.run_on_component_edgelist(edgelist, component_id)
        return res

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> list[str]:
        """Run the layout on a component.

        :param component: The component graph to run the analysis on.
        :return: Name of the parquet file containing the layout data.
        :raises TypeError: If the component is not a Graph or a LazyFrame.
        """
        results = []
        for algo in self._layout_algorithms:
            # TODO to get things working working with setting a different
            # k for pmgs w need to pass the weights from here, or change
            # the code in pixelator to have that as a parameter
            layout = component.layout_coordinates(
                algo,
                get_node_marker_matrix=False,
                **self._algorithm_kwargs,
            )
            layout["component"] = component_id
            layout["graph_projection"] = "full"
            layout["layout"] = algo
            results.append(layout)

        concatenated = pd.concat(results, axis=0).reset_index(drop=True)
        tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix=".parquet")
        concatenated.to_parquet(Path(tmp_file.name))
        return [tmp_file.name]

    def run_on_component_edgelist(
        self, component: pa.RecordBatch | pl.LazyFrame, component_id: str
    ) -> list[str]:
        """Run the layout on a component.

        :param component: The component to run the analysis on. Either a Graph or a LazyFrame.
        :param component_id: The id of the component.
        :return: Name of the parquet file containing the layout data.
        """
        if isinstance(component, pl.LazyFrame):
            graph = PNAGraph.from_edgelist(component)
        else:
            graph = PNAGraph.from_record_batches(component)
        result = self.run_on_component_graph(graph, component_id)

        return result

    def concatenate_data(self, data: Iterable[str]) -> list[str]:
        """Concatenate the data. Override this if you need custom concatenation behavior."""
        return list(itertools.chain.from_iterable(data))

    def add_to_pixel_file(self, data: list[str], pxl_file_target: PxlFile) -> None:
        """Add the data in the right place in the pxl_dataset."""
        paths = [Path(fname) for fname in data]
        with PixelFileWriter(pxl_file_target.path) as writer:
            writer.write_layouts(paths)

        for fname in data:
            os.remove(Path(fname))



"""Module contains classes and functions related to the configuration file for pixelator (assay settings).

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import itertools
import typing
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import importlib_resources
import semver

from pixelator.common.config.assay import Assay
from pixelator.common.config.panel import AntibodyPanel
from pixelator.common.exceptions import PixelatorBaseException
from pixelator.common.types import PathType

DNA_CHARS = {"A", "C", "G", "T"}

RangeType = typing.TypeVar(
    "RangeType", Tuple[int, int], Tuple[Optional[int], Optional[int]]
)


class PanelException(PixelatorBaseException):
    """Exception raised for failures to load a panel into the global configuration."""

    pass


class Config:
    """Class containing the pixelator configuration (assay settings)."""

    def __init__(
        self,
        assays: Optional[List[Assay]] = None,
        panels: Optional[List[AntibodyPanel]] = None,
    ) -> None:
        """Initialize the config object."""
        self.assays: Dict[str, Assay] = {}
        self.panels: typing.MutableMapping[str, List[AntibodyPanel]] = defaultdict(list)
        self.panel_aliases: Dict[str, str] = {}

        if assays is not None:
            self.assays.update({a.name: a for a in assays})

        if panels is not None:
            for p in panels:
                key = p.name if p.name is not None else str(p.filename)
                self.panels[key].append(p)

    def load_assay(self, path: PathType) -> None:
        """Load an assay from a yaml file."""
        assay = Assay.from_yaml(path)
        self.assays[assay.name] = assay

    def load_panel_file(self, path: PathType) -> None:
        """Load the panel file.

        :param path: The path to the panel file.
        :raises PanelException: If the panel alias already exists in the config.
        """
        panel = AntibodyPanel.from_csv(path)
        key = panel.name if panel.name is not None else str(panel.filename)
        self.panels[key].append(panel)

        # Enable panel lookup by aliases
        for alias in panel.aliases:
            if (alias in self.panel_aliases) and (key != self.panel_aliases[alias]):
                raise PanelException(
                    f'Panel alias "{alias}" already exists in the '
                    f'config for panel "{self.panel_aliases[alias]}".'
                    "If you provided your own panel file, please "
                    "ensure the panel name an aliases are unique "
                    "in the header of the file."
                )
                continue

            self.panel_aliases[alias] = key

    def load_assays(self, path: PathType):
        """Load all assays from a directory containing yaml files."""
        search_path = Path(path)

        yaml_files = list(
            itertools.chain(
                search_path.glob("*.yaml"),
                search_path.glob("*.yml"),
            )
        )

        for f in yaml_files:
            self.load_assay(f)

    def load_panels(self, path: PathType):
        """Load all panel files from a directory containing csv files."""
        search_path = Path(path)

        csv_files = list(search_path.glob("*.csv"))

        for f in csv_files:
            self.load_panel_file(f)

    def get_assay(self, assay_name: str) -> Optional[Assay]:
        """Get an assay by name."""
        return self.assays.get(assay_name)

    def list_panel_names(self, include_aliases: bool = False) -> List[str]:
        """Return a list of all panel names.

        :param include_aliases: Include panel aliases in the list
        :returns: A list of panel names
        """
        out = sorted(list(self.panels.keys()))

        if not include_aliases:
            return out

        out += sorted(list(self.panel_aliases.keys()))
        return out

    def get_panel(
        self,
        panel_name: str,
        version: Optional[str] = None,
        allow_aliases: bool = True,
    ) -> Optional[AntibodyPanel]:
        """Get a panel by name.

        :param panel_name: The name of the panel
        :param version: The optional version of a panel to return
        :param allow_aliases: Allow panel aliases to be used
        """
        panels_with_key = self.panels.get(panel_name)

        # Try to load using an alias if no name matches are found
        if panels_with_key is None and allow_aliases:
            panel_alias = self.panel_aliases.get(panel_name)
            if panel_alias is not None:
                panels_with_key = self.panels.get(panel_alias)

        if panels_with_key is None:
            return None

        def keyfunc(p):
            version = p.version
            if version is None:
                v = semver.Version.parse("0.0.0")

            v = semver.Version.parse(version)
            return v

        panels_with_key = sorted(panels_with_key, key=keyfunc, reverse=True)
        if version is None:
            return panels_with_key[0]

        selected_panel = next(
            (p for p in panels_with_key if p.version == version), None
        )
        return selected_panel


def load_assays_package(config: Config, package_name: str) -> Config:
    """Load default assays from a resources package.

    :param config: The config object to load assays into
    :param package_name: The name of the package to load assays from
    :return: The updated config object
    """
    # TODO: Consider switching to base importlib.resources after
    #       dropping python3.8 support.
    for resource in importlib_resources.files(package_name).iterdir():
        if resource.is_file():
            with importlib_resources.as_file(resource) as file_path:
                config.load_assay(file_path)

    return config


def load_panels_package(config: Config, package_name: str) -> Config:
    """Load default panels from a resources package.

    :param config: The config object to load panel files into
    :param package_name: The name of the package to load panels from
    :return: The updated config object
    """
    # TODO: Consider switching to base importlib.resources after
    #       dropping python3.8 support.
    for resource in importlib_resources.files(package_name).iterdir():
        if resource.is_file():
            with importlib_resources.as_file(resource) as file_path:
                config.load_panel_file(file_path)

    return config



"""Model for sample metadata.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
from collections import Counter
from pathlib import Path
from typing import Iterable

import pandas as pd
import pydantic

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

logger = logging.getLogger(__name__)


class SampleMetadataRecord(pydantic.BaseModel):
    """Model for sample metadata.

    :ivar sample_id: The sample id
    :ivar description: A description of the sample
    :ivar panel_version: Semantic version string of a panel file
    :ivar panel_name: The name of the panel
    """

    sample_id: str = pydantic.Field(description="The sample id")

    description: str = pydantic.Field(description="A description of the sample")

    panel_version: str = pydantic.Field(
        description="Semantic version string of a panel file"
    )

    panel_name: str = pydantic.Field(description="The name of the panel")


class SampleMetadata:
    """Metadata for a collection of samples."""

    def __init__(self, metadata: Iterable[SampleMetadataRecord] = ()):
        """Create a SampleMetadata object.

        :param metadata: An iterable of :class:`SampleMetadataRecord` objects.
        """
        self._data = list(metadata)
        self._lookup = {v.sample_id: v for v in self._data}
        self._validate_data()

    def get_by_id(self, sample_id: str) -> SampleMetadataRecord | None:
        """Retrieve a metadata record from a sample id.

        :param sample_id: The sample id to retrieve metadata for
        :returns: The metadata record for the sample id or None if not found
        """
        return self._lookup.get(sample_id)

    def _validate_data(self):
        """Check for duplicate sample ids."""
        if len(self._data) != len(self._lookup.keys()):
            id_counts = Counter((v.sample_id for v in self._data))
            for sample_id, count in id_counts.items():
                if count > 1:
                    raise ValueError(
                        f"Every sample must have a unique id: duplicate "
                        f"sample id: {sample_id}"
                    )

    @classmethod
    def from_csv(cls, p: Path | str) -> Self:
        """Create a SampleMetadata object from a csv file.

        :arg p: The path to the csv file.
        :returns: A SampleMetadata object.
        """
        logger.debug("Parsing metadata from %s", str(p))
        data = pd.read_csv(
            p,
            sep=",",
            index_col=0,
            dtype={
                "sample_id": str,
                "description": str,
                "panel_version": str,
                "panel_name": str,
            },
        )
        records = []

        for record_name, record in data.iterrows():
            records.append(SampleMetadataRecord(**record.to_dict()))

        return cls(records)



"""
Console script for pixelator (layout)

Copyright © 2024 Pixelgen Technologies AB.
"""

from typing import get_args

import click

from pixelator.common.graph.backends.protocol import SupportedLayoutAlgorithm
from pixelator.common.utils import (
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx import read
from pixelator.mpx.cli.common import logger, output_option
from pixelator.mpx.pixeldataset.precomputed_layouts import (
    generate_precomputed_layouts_for_components,
)
from pixelator.mpx.report.common import PixelatorWorkdir
from pixelator.mpx.report.models.layout import LayoutSampleReport


@click.command(
    "layout",
    short_help="compute graph layouts that can be used to visualize components",
    options_metavar="<options>",
)
@click.argument(
    "pxl_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="PXL",
)
@click.option(
    "--no-node-marker-counts",
    required=False,
    is_flag=True,
    default=False,
    help="Skip adding marker counts to the layout. Default: False.",
)
@click.option(
    "--layout-algorithm",
    required=False,
    multiple=True,
    default=["wpmds_3d"],
    help="Select a layout algorithm to use. This can be specified multiple times to compute multiple layouts. Default: pmds_3d",
    type=click.Choice(get_args(SupportedLayoutAlgorithm)),
)
@output_option
@click.pass_context
@timer
def layout(
    ctx,
    pxl_file,
    no_node_marker_counts,
    layout_algorithm,
    output,
):
    """
    Compute graph layouts that can be used to visualize components
    """
    log_step_start(
        "layout",
        input_files=pxl_file,
        no_node_marker_counts=no_node_marker_counts,
        layout_algorithm=layout_algorithm,
        output=output,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(pxl_file, allowed_extensions="pxl")

    # create output folder if it does not exist
    workdir = PixelatorWorkdir(output)
    layout_output_dir = workdir.stage_dir("layout")

    logger.info(f"Computing layout(s) for file {pxl_file}")

    clean_name = get_sample_name(pxl_file)
    write_parameters_file(
        ctx,
        layout_output_dir / f"{clean_name}.meta.json",
        command_path="pixelator single-cell-mpx layout",
    )

    pxl_dataset = read(pxl_file)
    pxl_dataset.precomputed_layouts = generate_precomputed_layouts_for_components(
        pxl_dataset,
        add_node_marker_counts=not no_node_marker_counts,
        layout_algorithms=layout_algorithm,
    )
    pxl_dataset.save(
        layout_output_dir / f"{clean_name}.layout.dataset.pxl", force_overwrite=True
    )

    metrics_file = layout_output_dir / f"{clean_name}.report.json"
    report = LayoutSampleReport(
        sample_id=clean_name,
    )
    report.write_json_file(metrics_file, indent=4)



"""Code adapted from cutadapt.

Modifications Copyright © 2025 Pixelgen Technologies AB.
Under the same license terms as the original code.

Original copyright notice:
Copyright (c) 2010 Marcel Martin <marcel.martin@scilifelab.se> and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import io
import logging
import multiprocessing
import os
import sys
import traceback
import typing
from abc import ABC, abstractmethod
from contextlib import ExitStack
from multiprocessing.connection import Connection
from typing import (
    TYPE_CHECKING,
    Any,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
)

import dnaio
from cutadapt.files import (
    FileFormat,
    InputFiles,
    InputPaths,
    OutputFiles,
    ProxyWriter,
    detect_file_format,
    xopen_rb_raise_limit,
)
from cutadapt.pipeline import Pipeline
from cutadapt.report import Statistics
from cutadapt.utils import Progress

logger = logging.getLogger()

mpctx = multiprocessing.get_context("spawn")

# See https://github.com/python/typeshed/issues/9860
if TYPE_CHECKING:
    mpctx_Process = multiprocessing.Process  # pragma: no cover
else:
    mpctx_Process = mpctx.Process


class ReaderProcess(mpctx_Process):
    """Read chunks of FASTA or FASTQ data (single-end or paired) and send them to a worker.

    The reader repeatedly

    - reads a chunk from the file(s)
    - reads a worker index from the Queue
    - sends the chunk to connections[index]

    and finally sends the stop token -1 ("poison pills") to all connections.
    """

    def __init__(
        self,
        *paths: str,
        file_format_connection: Connection,
        connections: Sequence[Connection],
        queue: multiprocessing.Queue,
        buffer_size: int,
        stdin_fd,
    ):
        """Initialize a ReaderProcess.

        :param params paths: path to input files
        :param connections: a list of Connection objects, one for each worker.
        :param queue: a Queue of worker indices. A worker writes its own index into this
        :param queue to notify the reader that it is ready to receive more data.
        :param buffer_size:
        :param stdin_fd:

        .. note::
            This expects the paths to the input files as strings because these can be pickled
            while file-like objects such as BufferedReader cannot. When using multiprocessing with
            the "spawn" method, which is the default method on macOS, function arguments must be
            picklable.
        """
        super().__init__()
        if len(paths) > 2:
            raise ValueError("Reading from more than two files currently not supported")
        if not paths:
            raise ValueError("Must provide at least one file")
        self._paths = paths
        self._file_format_connection = file_format_connection
        self.connections = connections
        self.queue = queue
        self.buffer_size = buffer_size
        self.stdin_fd = stdin_fd

    def run(self):
        """Read chunks of data from the input files and send them to workers."""
        if self.stdin_fd != -1:
            sys.stdin.close()
            sys.stdin = os.fdopen(self.stdin_fd)
        try:
            with ExitStack() as stack:
                try:
                    files = [
                        stack.enter_context(xopen_rb_raise_limit(path))
                        for path in self._paths
                    ]
                    file_format = detect_file_format(files[0])
                except Exception as e:
                    self._file_format_connection.send(-2)
                    self._file_format_connection.send((e, traceback.format_exc()))
                    raise
                self._file_format_connection.send(file_format)
                for index, chunks in enumerate(self._read_chunks(*files)):
                    self.send_to_worker(index, *chunks)
            self.shutdown()
        except Exception as e:
            # TODO better send this to a common "something went wrong" Queue
            # This code is rarely executed because there is little that can go wrong
            # splitting up the input into chunks. FASTQ/FASTA parsing problems
            # are caught within the workers.
            for connection in self.connections:
                connection.send(-2)
                connection.send((e, traceback.format_exc()))

    def _read_chunks(self, *files) -> Iterator[Tuple[memoryview, ...]]:
        if len(files) == 1:
            for chunk in dnaio.read_chunks(files[0], self.buffer_size):
                yield (chunk,)
        elif len(files) == 2:
            for chunks in dnaio.read_paired_chunks(
                files[0], files[1], self.buffer_size
            ):
                yield chunks
        else:
            raise NotImplementedError

    def send_to_worker(self, chunk_index, chunk1, chunk2=None):
        """Send a chunk of data to a worker.

        :param chunk_index: The index of the chunk.
        :param chunk1: The first chunk of data.
        :param chunk2: The second chunk of data (if paired-end data).
        """
        worker_index = self.queue.get()
        connection = self.connections[worker_index]
        connection.send(chunk_index)
        connection.send_bytes(chunk1)
        if chunk2 is not None:
            connection.send_bytes(chunk2)

    def shutdown(self):
        """Notify all workers to shut down."""
        # Send `-1` poison pills to all workers
        for _ in range(len(self.connections)):
            worker_index = self.queue.get()
            self.connections[worker_index].send(-1)


class WorkerProcess(mpctx_Process):
    """A custom multiprocessing Proces for running a pipeline on chunks of data.

    The worker repeatedly reads chunks of data from the read_pipe, runs the pipeline on it
    and sends the processed chunks to the write_pipe.

    To notify the reader process that it wants data, it puts its own identifier into the
    need_work_queue before attempting to read data from the read_pipe.
    """

    def __init__(
        self,
        id_: int,
        pipeline: Pipeline,
        inpaths: InputPaths,
        proxy_files: List[ProxyWriter],
        read_pipe: Connection,
        write_pipe: Connection,
        need_work_queue: multiprocessing.Queue,
        statistics_class: Type[Statistics] = Statistics,
    ):
        """Initialize a WorkerProcess."""
        super().__init__()
        self._id = id_
        self._pipeline = pipeline
        self._n_input_files = len(inpaths.paths)
        self._interleaved_input = inpaths.interleaved
        self._read_pipe = read_pipe
        self._write_pipe = write_pipe
        self._need_work_queue = need_work_queue
        self._proxy_files = proxy_files
        self._statistics_class = statistics_class

    def run(self):
        """Read chunks of data from the reader, process them and send them to the writer."""
        try:
            stats = self._statistics_class()

            while True:
                # Notify reader that we need data
                self._need_work_queue.put(self._id)
                chunk_index = self._read_pipe.recv()
                if chunk_index == -1:
                    # reader is done
                    break
                elif chunk_index == -2:
                    # An exception has occurred in the reader
                    e, tb_str = self._read_pipe.recv()
                    logger.error("%s", tb_str)
                    raise e

                # Receive blocks of data from the reader
                files = [
                    io.BytesIO(self._read_pipe.recv_bytes())
                    for _ in range(self._n_input_files)
                ]
                # Pass the data to the pipeline as a file like object
                infiles = InputFiles(*files, interleaved=self._interleaved_input)
                (n, bp1, bp2) = self._pipeline.process_reads(infiles)
                stats += self._statistics_class().collect(n, bp1, bp2, [], [])
                self._send_outfiles(chunk_index, n)

            stats_instance = self._statistics_class()
            stats += stats_instance.collect(
                0,
                0,
                0 if self._pipeline.paired else None,
                self._pipeline._modifiers,
                self._pipeline._steps,
                set_paired_to_none=self._pipeline.paired is None,
            )
            self._write_pipe.send(-1)
            self._write_pipe.send(stats)

        except Exception as e:
            self._write_pipe.send(-2)
            self._write_pipe.send((e, traceback.format_exc()))

    def _send_outfiles(self, chunk_index: int, n_reads: int):
        self._write_pipe.send(chunk_index)
        self._write_pipe.send(n_reads)
        for pf in self._proxy_files:
            for chunk in pf.drain():
                self._write_pipe.send_bytes(chunk)


class OrderedChunkWriter:
    """Reorder output chunks according to input order.

    We may receive chunks of processed data from worker processes
    in any order. This class writes them to an output file in
    the correct order.
    """

    def __init__(self, outfile):
        """Initialize an OrderedChunkWriter."""
        self._chunks = dict()
        self._current_index = 0
        self._outfile = outfile

    def write(self, data: bytes, index: int):
        """Write a chunk of data to the output handler.

        :param data: The data to write.
        :param index: An integer indicating the order of the data.
        """
        self._chunks[index] = data
        while self._current_index in self._chunks:
            self._outfile.write(self._chunks[self._current_index])
            del self._chunks[self._current_index]
            self._current_index += 1

    def wrote_everything(self):
        """Return True if all chunks have been written."""
        return not self._chunks


class PipelineRunner(ABC):
    """A read processing pipeline."""

    @abstractmethod
    def run(
        self, pipeline: Pipeline, progress: Progress, outfiles: OutputFiles
    ) -> Statistics:
        """Run a pipeline on this runner.

        :param pipeline: The pipeline to run.
        :param progress: Use an object that supports .update() and .close() such
            as DummyProgress, cutadapt.utils.Progress or a tqdm instance.
        :param outfiles: The output files.
        """

    @abstractmethod
    def close(self):
        """Close all open file-descriptors."""
        pass

    @abstractmethod
    def input_file_format(self) -> FileFormat:
        """Return the input file format of the input file(s)."""
        pass

    def __enter__(self):
        """Initialize the runner."""
        return self

    def __exit__(self, *args):
        """Close the runner."""
        self.close()


class WorkerException(Exception):
    """An exception that occurred in a worker process.

    Embeds the original exception and the traceback string.
    """

    def __init__(self, wrapped_exception, tb_str):
        """Initialize a WorkerException."""
        self.e = wrapped_exception
        self.tb_str = tb_str

    def __str__(self):
        """Return a string representation of the exception."""
        return f"{self.e}\nwith traceback:\n{self.tb_str}"


StatisticsClass = typing.TypeVar("StatisticsClass", bound=Statistics)


class ParallelPipelineRunner(PipelineRunner, typing.Generic[StatisticsClass]):
    """Run a Pipeline in parallel.

    - At construction, a reader process is spawned.
    - When run() is called, as many worker processes as requested are spawned.
    - In the main process, results are written to the output files in the correct
      order, and statistics are aggregated.

    If a worker needs work, it puts its own index into a Queue() (_need_work_queue).
    The reader process listens on this queue and sends the raw data to the
    worker that has requested work. For sending the data from reader to worker,
    a Connection() is used. There is one such connection for each worker (self._pipes).

    For sending the processed data from the worker to the main process, there
    is a second set of connections, again one for each worker.

    When the reader is finished, it sends 'poison pills' to all workers.
    When a worker receives this, it sends a poison pill to the main process,
    followed by a Statistics object that contains statistics about all the reads
    processed by that worker.
    """

    def __init__(
        self,
        inpaths: InputPaths,
        n_workers: int,
        buffer_size: Optional[int] = None,
        statistics_class: StatisticsClass = Statistics,
    ):
        """Initialize a ParallelPipelineRunner.

        :param inpaths: The input files.
        :param n_workers: The number of worker processes to use.
        :param buffer_size: The size of the buffer used for reading the input files.
        :param statistics_class: The class to use for collecting statistics.
        """
        self._n_workers = n_workers
        self._need_work_queue: multiprocessing.Queue = mpctx.Queue()
        self._buffer_size = 4 * 1024**2 if buffer_size is None else buffer_size
        self._inpaths = inpaths
        # the workers read from these connections
        connections = [mpctx.Pipe(duplex=False) for _ in range(self._n_workers)]
        self._connections, connw = zip(*connections)
        try:
            fileno = sys.stdin.fileno()
        except io.UnsupportedOperation:
            # This happens during tests: pytest sets sys.stdin to an object
            # that does not have a file descriptor.
            fileno = -1

        file_format_connection_r, file_format_connection_w = mpctx.Pipe(duplex=False)
        self._reader_process = ReaderProcess(
            *inpaths.paths,
            file_format_connection=file_format_connection_w,
            connections=connw,
            queue=self._need_work_queue,
            buffer_size=self._buffer_size,
            stdin_fd=fileno,
        )
        self._reader_process.daemon = True
        self._reader_process.start()
        self._input_file_format = self._try_receive(file_format_connection_r)
        self._statistics_class = statistics_class

    def _start_workers(
        self, pipeline, proxy_files
    ) -> Tuple[List[WorkerProcess], List[Connection]]:
        workers = []
        connections = []
        for index in range(self._n_workers):
            conn_r, conn_w = mpctx.Pipe(duplex=False)
            connections.append(conn_r)
            worker = WorkerProcess(
                index,
                pipeline,
                self._inpaths,
                proxy_files,
                self._connections[index],
                conn_w,
                self._need_work_queue,
                statistics_class=self._statistics_class,
            )
            worker.daemon = True
            worker.start()
            workers.append(worker)
        return workers, connections

    def run(self, pipeline, progress, outfiles: OutputFiles) -> StatisticsClass:
        """Run the pipeline on the input files.

        :param pipeline: The pipeline to run.
        :param progress: A progress object.
        :param outfiles: The output files.
        :return: The statistics object.
        """
        workers, connections = self._start_workers(pipeline, outfiles.proxy_files())
        chunk_writers = []

        for f in outfiles.binary_files():
            chunk_writers.append(OrderedChunkWriter(f))

        stats = self._statistics_class()
        while connections:
            ready_connections: List[Any] = multiprocessing.connection.wait(connections)
            for connection in ready_connections:
                chunk_index: int = self._try_receive(connection)
                if chunk_index == -1:
                    # the worker is done
                    cur_stats = self._try_receive(connection)
                    stats += cur_stats
                    connections.remove(connection)
                    continue

                number_of_reads: int = self._try_receive(connection)
                progress.update(number_of_reads)

                for writer in chunk_writers:
                    data = connection.recv_bytes()
                    writer.write(data, chunk_index)

        for writer in chunk_writers:
            assert writer.wrote_everything()

        for w in workers:
            w.join()

        self._reader_process.join()
        progress.close()
        return stats

    @staticmethod
    def _try_receive(connection):
        """Try to receive data over `connection` and return it.

        If an exception was received, raise it.
        """
        result = connection.recv()
        if result == -2:
            # An exception has occurred on the other end
            e, tb_str = connection.recv()
            # The other end does not send an actual traceback object because these are
            # not picklable, but a string representation.
            logger.debug("%s", tb_str)
            for child in multiprocessing.active_children():
                child.terminate()
            raise WorkerException(e, tb_str)
        return result

    def close(self) -> None:
        """Close all open file-descriptors."""
        pass

    def input_file_format(self) -> FileFormat:
        """Return the input file format of the input file(s)."""
        return self._input_file_format


class SerialPipelineRunner(PipelineRunner):
    """Run a Pipeline on a single core."""

    def __init__(
        self,
        infiles: InputFiles,
        statistics_class: Type[Statistics] = Statistics,
    ):
        """Initialize a SerialPipelineRunner.

        :param infiles: The input files.
        :param statistics_class: The class to use for collecting statistics.
        """
        self._infiles = infiles
        self._input_file_format = infiles
        self._statistics_class = statistics_class

    def run(
        self, pipeline: Pipeline, progress: Progress, outfiles: OutputFiles
    ) -> Statistics:
        """Run the pipeline on the input files.

        :param pipeline: The pipeline to run.
        :param progress: A progress object.
        :param outfiles: The output files.
        :return: The statistics object.
        """
        (n, total1_bp, total2_bp) = pipeline.process_reads(
            self._infiles, progress=progress
        )
        if progress is not None:
            progress.close()

        modifiers = getattr(pipeline, "_modifiers", None)
        assert modifiers is not None
        return self._statistics_class().collect(
            n, total1_bp, total2_bp, modifiers, pipeline._steps
        )  # type: ignore[attr-defined]

    def close(self):
        """Close all open file-descriptors from the `infiles`."""
        self._infiles.close()

    def input_file_format(self) -> FileFormat:
        """Return the detected file format of the input file(s)."""
        return detect_file_format(self._infiles._files[0])


def make_runner(
    inpaths: InputPaths,
    cores: int,
    buffer_size: Optional[int] = None,
    statistics_class: Type[Statistics] = Statistics,
) -> PipelineRunner:
    """Run a pipeline.

    This uses a SerialPipelineRunner if cores is 1 and a ParallelPipelineRunner otherwise.

    :param inpaths: The input files.
    :param cores: The number of cores to run the pipeline on (this is actually the number of worker
                  processes, there will be one extra process for reading the input file(s))
    :param buffer_size: Forwarded to `ParallelPipelineRunner()`. Ignored if cores is 1.
    """
    runner: PipelineRunner
    if cores > 1:
        runner = ParallelPipelineRunner(
            inpaths,
            n_workers=cores,
            buffer_size=buffer_size,
            statistics_class=statistics_class,
        )
    else:
        runner = SerialPipelineRunner(
            inpaths.open(),
            statistics_class=statistics_class,
        )

    return runner



"""Copyright © 2023 Pixelgen Technologies AB."""

from pixelator.mpx.collapse.process import (
    collapse_fastq,
    collapse_sequences_adjacency,
    collapse_sequences_unique,
)

__all__ = [
    "collapse_fastq",
    "collapse_sequences_adjacency",
    "collapse_sequences_unique",
]



"""Console script for pixelator (amplicon)

Copyright © 2022 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging

import click

from pixelator.common.utils import (
    create_output_stage_dir,
    get_extension,
    get_read_sample_name,
    is_read_file,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.mpx.amplicon import amplicon_fastq
from pixelator.mpx.cli.common import (
    design_option,
    logger,
    output_option,
)


@click.command(
    "amplicon",
    short_help=("process diverse raw pixel data (FASTQ) formats into common amplicon"),
    options_metavar="<options>",
)
@click.argument(
    "fastq_1",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="FASTQ_1",
)
@click.argument(
    "fastq_2",
    nargs=1,
    required=False,
    type=click.Path(exists=True),
    metavar="FASTQ_2",
)
@click.option(
    "--sample-name",
    default=None,
    show_default=False,
    type=click.STRING,
    help=(
        "Override the basename of the output fastq file. "
        "Default is the basename of the first input file "
        "without extension and read 1 identifier."
    ),
)
@click.option(
    "--skip-input-checks",
    default=False,
    is_flag=True,
    type=click.BOOL,
    help="Skip all check on the filename of input fastq files.",
)
@output_option
@design_option
@click.pass_context
@timer
def amplicon(
    ctx,
    fastq_1: str,
    fastq_2: str | None,
    sample_name: str | None,
    skip_input_checks: bool,
    output: str,
    design: str,
):
    """
    Process diverse raw pixel data (FASTQ) formats into common amplicon
    """
    # log input parameters

    error_level = logging.WARNING if skip_input_checks else logging.ERROR

    log_step_start(
        "amplicon",
        fastq1=fastq_1,
        fastq_2=fastq_2,
        output=output,
        design=design,
    )

    # some basic sanity check on the input files
    fastq_inputs = [fastq_1] + [fastq_2] if fastq_2 else []
    sanity_check_inputs(fastq_inputs, allowed_extensions=("fastq.gz", "fq.gz"))

    # create output folder if it does not exist
    amplicon_output = create_output_stage_dir(output, "amplicon")

    # Some checks on the input files
    # - check if there are read 1 and read2 identifiers in the filename
    # - check if the sample name is the same for read1 and read2
    if not is_read_file(fastq_1, "r1"):
        msg = "Read 1 file does not contain a recognised read 1 suffix."
        logger.log(level=error_level, msg=msg)
        if not skip_input_checks:
            ctx.exit(1)

    if fastq_2 and not is_read_file(fastq_2, "r2"):
        msg = "Read 2 file does not contain a recognised read 2 suffix."
        logger.log(level=error_level, msg=msg)
        if not skip_input_checks:
            ctx.exit(1)

    r1_sample_name = get_read_sample_name(fastq_1)
    r2_sample_name = get_read_sample_name(fastq_2) if fastq_2 else None

    if fastq_2 and r1_sample_name != r2_sample_name:
        msg = (
            f"The sample name for read1 and read2 is different:\n"
            f'"{r1_sample_name}" vs "{r2_sample_name}"\n'
            "Did you pass the correct files?"
        )
        logger.log(level=error_level, msg=msg)
        if not skip_input_checks:
            ctx.exit(1)

    sample_name = sample_name or r1_sample_name
    extension = get_extension(fastq_1)
    output_file = amplicon_output / f"{sample_name}.merged.{extension}"
    json_file = amplicon_output / f"{sample_name}.report.json"

    write_parameters_file(
        ctx,
        amplicon_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-mpx amplicon",
    )

    msg = f"Creating amplicon for {','.join(str(p) for p in fastq_inputs)}"
    logger.info(msg)

    amplicon_fastq(
        inputs=fastq_inputs,
        design=design,
        metrics=json_file,
        sample_id=sample_name,
        output=str(output_file),
    )



"""Logging setup for pixelator.

Copyright © 2022 Pixelgen Technologies AB.
"""

import logging
import pickle
import socketserver
import struct
import sys
import threading
import time
import traceback
import typing
import warnings
from logging.handlers import DEFAULT_TCP_LOGGING_PORT, SocketHandler
from pathlib import Path

import click
from numba import NumbaDeprecationWarning

from pixelator.common.types import PathType

# Silence deprecation warnings
warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.filterwarnings(
    "ignore",
    message="geopandas not available. Some functionality will be disabled.",
    category=UserWarning,
)


# Disable matplot lib debug logs (that will clog all debug logs)
logging.getLogger("matplotlib.font_manager").setLevel(logging.ERROR)
logging.getLogger("matplotlib.ticker").setLevel(logging.ERROR)
logging.getLogger("numba").setLevel(logging.ERROR)


# ------------------------------------------------------------
# Click logging
# ------------------------------------------------------------


class StyleDict(typing.TypedDict):
    """Style dictionary for kwargs to `click.style`."""

    fg: str


class ColorFormatter(logging.Formatter):
    """Click formatter with colored levels."""

    colors: dict[str, StyleDict] = {
        "debug": StyleDict(fg="blue"),
        "info": StyleDict(fg="green"),
        "warning": StyleDict(fg="yellow"),
        "error": StyleDict(fg="red"),
        "exception": StyleDict(fg="red"),
        "critical": StyleDict(fg="red"),
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format a record with colored level.

        :param record: The record to format.
        :returns str: A formatted log record.
        """
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.getMessage()
            if level in self.colors:
                timestamp = self.formatTime(record, self.datefmt)
                colored_level = click.style(
                    f"{level.upper():<10}", **self.colors[level]
                )
                prefix = f"{timestamp} [{colored_level}]  "
                msg = "\n".join(prefix + x for x in msg.splitlines())
            return msg
        return logging.Formatter.format(self, record)


class DefaultCliFormatter(logging.Formatter):
    """Click formatter with colored levels."""

    def format(self, record: logging.LogRecord) -> str:
        """Format a record for CLI output."""
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.getMessage()

            if level == "info":
                return msg

            return f"{level.upper()}: {msg}"
        return logging.Formatter.format(self, record)


class ClickHandler(logging.Handler):
    """Click logging handler.

    Messages are forwarded to stdout using `click.echo`.
    """

    def __init__(self, level: int = 0, use_stderr: bool = True):
        """Initialize the click handler.

        :param level: The logging level.
        :param use_stderr: Log to sys.stderr instead of sys.stdout.
        """
        super().__init__(level=level)
        self._use_stderr = use_stderr

    def emit(self, record: logging.LogRecord) -> None:
        """Do whatever it takes to actually log the specified logging record.

        :param record: The record to log.
        """
        try:
            msg = self.format(record)
            click.echo(msg, err=self._use_stderr)
        except Exception:
            self.handleError(record)


LOCALHOST = "localhost"


class LoggingSetup:
    """Logging setup for pixelator.

    This class is used to set up logging for pixelator. The reason we need this somewhat
    involved setup is to be able to deal with

    All messages are passed to a separate process that handles the logging to a file.
    """

    def __init__(
        self, log_file: PathType | None = None, verbose: bool = False, logger=None
    ):
        """Initialize the logging setup.

        :param log_file: the filename of the log output
        :param verbose: enable verbose logging and console output
        :param logger: the logger to configure, default is the root logger
        """
        self.log_file = Path(log_file) if log_file is not None else None
        self.verbose = verbose
        self._root_logger = logger or logging.getLogger()
        self._server = None
        self._server_thread = None

    def _shutdown_listener(self) -> None:
        """Send the stop token to the logging process and wait for it to finish."""
        if self._server:
            self._server.shutdown()
            self._server.server_close()
        if self._server_thread:
            self._server_thread.join(timeout=1)

    def initialize(self):
        """Configure logging and start the listener process."""
        self._server = LogRecordSocketReceiver(
            host=LOCALHOST,
            # Setting the port to 0 asks the OS to assign a free port
            port=0,
            handler=LogRecordStreamHandler,
            log_file=self.log_file,
            console_log_formatter=(
                ColorFormatter(datefmt="%Y-%m-%d %H:%M:%S")
                if self.verbose
                else DefaultCliFormatter()
            ),
        )
        _, port = self._server.server_address
        self.port = port
        self._server_thread = threading.Thread(
            target=self._server.serve_forever, name="pixelator-log-listener"
        )
        self._server_thread.daemon = True
        self._server_thread.start()

        socker_handler = SocketHandler(LOCALHOST, port)
        self._root_logger.addHandler(socker_handler)
        self._root_logger.setLevel(logging.DEBUG if self.verbose else logging.INFO)

    @property
    def log_level(self):
        """Return the current log level."""
        return self._root_logger.level

    def __enter__(self):
        """Enter the context manager.

        This will initialize the logging setup.
        """
        self.initialize()
        return self

    def close(self):
        """Close down the logging setup."""
        self._shutdown_listener()

    def __exit__(self, exc_type, exc_value, traceback_obj):
        """Exit the context manager.

        This will shut down the logging process if needed.
        """

        def log_exception(exc_type, exc_value, traceback_obj):
            if issubclass(exc_type, click.exceptions.UsageError) or issubclass(
                exc_type, click.exceptions.Exit
            ):
                # Click exceptions are dealt with separately
                # and thus we ignore them here.
                return False

            if issubclass(exc_type, SystemExit):
                # SystemExit is raised when the application has been explicitly
                # directed to exit, so we don't what a trace dumped for that.
                return False

            self._root_logger.critical(
                "Unhandled exception of type: {}".format(exc_type.__name__)
            )
            self._root_logger.critical("Exception message was: {}".format(exc_value))
            for item in traceback.format_exception(exc_type, exc_value, traceback_obj):
                for line in item.splitlines():
                    self._root_logger.critical(line)

        try:
            # It seems that sometimes the current exception is not passed
            # to the __exit__ method. This allows us to check if there
            # is any raised exception and make sure this gets added to
            # the log.
            if exc_type is None:
                exc_type, exc_value, traceback_obj = sys.exc_info()
                if exc_type is not None:
                    log_exception(exc_type, exc_value, traceback_obj)
            else:
                log_exception(exc_type, exc_value, traceback_obj)
        finally:
            self.close()

        # Reraise exception higher up the stack
        return False


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This will not filter log records, so make sure only to send
    messages that you actually want in the log to it.
    """

    timeout = 1

    def handle(self):
        """Handle requests."""
        while True:
            try:
                chunk = self.connection.recv(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack(">L", chunk)[0]
                chunk = self.connection.recv(slen)
                while len(chunk) < slen:
                    chunk = chunk + self.connection.recv(slen - len(chunk))
                obj = pickle.loads(chunk)
                record = logging.makeLogRecord(obj)
                self.handle_log_record(record)
            except TimeoutError:
                if self.server.is_shutting_down.is_set():
                    break

    def handle_log_record(self, record):
        """Handle a log record."""
        logger = logging.getLogger(LogRecordSocketReceiver.LISTENER_LOGGER)
        logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """A simple TCP server that will listen for log records from on a TCP-socket.

    The caller is responsible for making sure that the server is properly shutdown
    by calling:

    ```
    server.shutdown()
    server.server_close()
    ```

    Please note that this must be called while `server_forever` is running from
    a separate thread, otherwise it will deadlock.

    """

    LISTENER_LOGGER = "pixelator-logger-listener"

    allow_reuse_address = True

    def __init__(
        self,
        host=LOCALHOST,
        port=DEFAULT_TCP_LOGGING_PORT,
        handler=LogRecordStreamHandler,
        log_file=None,
        console_log_formatter=DefaultCliFormatter(),
    ):
        """Initialize the log record socket receiver."""
        self.timeout = 0.1
        self.log_file = log_file
        self._console_log_formatter = console_log_formatter
        # Use this threading.Event to make sure that the handler
        # shuts down properly when the server is closed.
        self.is_shutting_down = threading.Event()
        self._configure_logger()
        super().__init__((host, port), handler)

    def _configure_logger(self):
        """Configure the listener logger.

        This will add a file logger if a log file has been specified,
        and it will always output to the console (with the provided
        formatter).
        """
        logger = logging.getLogger(self.LISTENER_LOGGER)
        logger.setLevel(logging.DEBUG)
        logger.propagate = False

        # Reset the handlers to make sure they are not duplicated
        # if multiple server instances are created.
        logger.handlers = []

        handlers = []
        if self.log_file:
            handler = logging.FileHandler(str(self.log_file), mode="w")
            formatter = logging.Formatter(
                "%(asctime)s %(processName)-10s %(name)s %(levelname)-8s %(message)s"
            )
            handler.setFormatter(formatter)
            handlers.append(handler)

        console_handler = ClickHandler()
        console_handler.setFormatter(self._console_log_formatter)
        handlers.append(console_handler)

        logger.handlers = handlers

    def server_close(self) -> None:
        """Close the server."""
        # Ensure any outstanding requests are handled before shutting down
        self.is_shutting_down.set()
        self.handle_request()

        # TODO If we don't sleep here the tests get very flakey
        # And I haven't been able to figure out any other way to fix it.
        time.sleep(0.01)

        logger = logging.getLogger(self.LISTENER_LOGGER)
        for handler in logger.handlers:
            handler.close()



"""Pre-computed layouts for components.

This module contains the functionality to generate and work with pre-computed
graph layouts. The primary purpose of this is to allow components to be
visualized quickly in downstream analysis, since layout computation
is a relatively computationally expensive operation.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional, Protocol

import pandas as pd
import polars as pl
import pyarrow.dataset as ds
import pyarrow.parquet as pq

from pixelator.common.exceptions import PixelatorBaseException
from pixelator.common.graph.backends.protocol import SupportedLayoutAlgorithm
from pixelator.common.marks import experimental
from pixelator.common.utils import batched, flatten, get_pool_executor
from pixelator.mpx.graph import Graph

if TYPE_CHECKING:
    from pixelator.mpx.pixeldataset import PixelDataset

import logging

logger = logging.getLogger(__name__)


def _write_parquet(frame: pl.LazyFrame, path: Path, partitioning: list[str]) -> None:
    table = frame.collect().to_arrow()
    file_options = ds.ParquetFileFormat().make_write_options(
        compression="zstd",
    )
    ds.write_dataset(
        table,
        path,
        format="parquet",
        partitioning_flavor="hive",
        partitioning=partitioning,
        file_options=file_options,
        existing_data_behavior="overwrite_or_ignore",
    )


class PreComputedLayoutsEmpty(PixelatorBaseException):
    """Raised when trying to access an empty PreComputedLayouts instance."""


class _DataProvider(Protocol):
    def is_empty(self) -> bool: ...

    def to_df(self, columns: list[str] | None = None) -> pd.DataFrame: ...

    def lazy(self): ...

    def unique_components(self): ...

    def filter(
        self,
        component_ids: str | set[str] | None = None,
        graph_projections: str | set[str] | None = None,
        layout_methods: str | set[str] | None = None,
    ) -> pl.LazyFrame | list[pl.LazyFrame]: ...

    def write_parquet(self, path: Path, partitioning: list[str]) -> None:
        """Write a parquet file to the provided path."""
        ...


class _EmptyDataProvider(_DataProvider):
    def __init__(self) -> None:
        # This class needs no parameters
        pass

    def write_parquet(self, path: Path, partitioning: list[str]) -> None:
        """Write a parquet file to the provided path."""
        return

    def is_empty(self) -> bool:
        return True

    def to_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        raise PreComputedLayoutsEmpty("No layouts available")

    def lazy(self):
        raise PreComputedLayoutsEmpty("No layouts available")

    def unique_components(self):
        return {}

    def filter(
        self,
        component_ids: str | set[str] | None = None,
        graph_projections: str | set[str] | None = None,
        layout_methods: str | set[str] | None = None,
    ) -> list[pl.LazyFrame]:
        raise PreComputedLayoutsEmpty("No layouts available")


class _SingleFrameDataProvider(_DataProvider):
    def __init__(self, lazy_frame: pl.LazyFrame) -> None:
        self._lazy_frame = lazy_frame

    def is_empty(self) -> bool:
        return self.lazy().select(pl.col("component")).first().collect().is_empty()

    def to_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        if columns:
            return self._lazy_frame.select(columns).collect().to_pandas()
        return self.lazy().collect().to_pandas()

    def write_parquet(self, path: Path, partitioning: list[str]) -> None:
        """Write a parquet file to the provided path."""
        _write_parquet(self.lazy(), path, partitioning)

    def lazy(self):
        return self._lazy_frame

    def unique_components(self):
        return set(
            self.lazy()
            .select(pl.col("component").unique())
            .collect()
            .get_column("component")
        )

    def filter(
        self,
        component_ids: str | set[str] | None = None,
        graph_projections: str | set[str] | None = None,
        layout_methods: str | set[str] | None = None,
    ) -> pl.LazyFrame:
        component_ids = PreComputedLayouts._convert_to_set(component_ids)
        graph_projections = PreComputedLayouts._convert_to_set(graph_projections)
        layout_methods = PreComputedLayouts._convert_to_set(layout_methods)

        layouts_lazy_filtered = self.lazy()

        if component_ids:
            layouts_lazy_filtered = layouts_lazy_filtered.filter(
                pl.col("component").is_in(component_ids)
            )

        if graph_projections:
            layouts_lazy_filtered = layouts_lazy_filtered.filter(
                pl.col("graph_projection").is_in(graph_projections)
            )

        if layout_methods:
            layouts_lazy_filtered = layouts_lazy_filtered.filter(
                pl.col("layout").is_in(layout_methods)
            )

        return layouts_lazy_filtered


class _MultiFrameDataProvider(_DataProvider):
    def __init__(self, lazy_frames: Iterable[pl.LazyFrame]) -> None:
        self._lazy_frames = list(lazy_frames)

    def is_empty(self) -> bool:
        return all(
            frame.select(pl.col("component")).first().collect().is_empty()
            # We don't want to use `lazy()` here to skip
            # concatenating which is an unnecessary operation
            # for this
            for frame in self._lazy_frames
        )

    def to_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        if columns:
            return self.lazy().select(columns).collect().to_pandas()
        return self.lazy().collect().to_pandas()

    def lazy(self):
        try:
            return pl.concat(self._lazy_frames, how="diagonal")
        except ValueError:
            raise PreComputedLayoutsEmpty(f"No layouts available: {self._lazy_frames}")

    def unique_components(self):
        return set(
            flatten(
                frame.select(pl.col("component").unique())
                .collect()
                .get_column("component")
                .to_list()
                for frame in self._lazy_frames
            )
        )

    def filter(
        self,
        component_ids: str | set[str] | None = None,
        graph_projections: str | set[str] | None = None,
        layout_methods: str | set[str] | None = None,
    ) -> list[pl.LazyFrame]:
        component_ids = PreComputedLayouts._convert_to_set(component_ids)
        graph_projections = PreComputedLayouts._convert_to_set(graph_projections)
        layout_methods = PreComputedLayouts._convert_to_set(layout_methods)

        def data():
            for frame in self._lazy_frames:
                if component_ids:
                    frame = frame.filter(pl.col("component").is_in(component_ids))

                if graph_projections:
                    frame = frame.filter(
                        pl.col("graph_projection").is_in(graph_projections)
                    )

                if layout_methods:
                    frame = frame.filter(pl.col("layout").is_in(layout_methods))

                if not frame.select(pl.col("component")).first().collect().is_empty():
                    yield frame

        return list(data())

    def write_parquet(self, path: Path, partitioning: list[str]) -> None:
        """Write a parquet file to the provided path."""
        for frame in self._lazy_frames:
            _write_parquet(frame, path, partitioning)


class PreComputedLayouts:
    """Pre-computed layouts for a set of graphs, per component."""

    DEFAULT_PARTITIONING = [
        "graph_projection",
        "layout",
        "component",
    ]

    def __init__(
        self,
        layouts_lazy: pl.LazyFrame | Iterable[pl.LazyFrame] | None,
        partitioning: Optional[list[str]] = None,
    ) -> None:
        """Initialize the PreComputedLayouts instance."""
        if layouts_lazy is None:
            self._data_provider: _DataProvider = _EmptyDataProvider()
        elif isinstance(layouts_lazy, pl.LazyFrame):
            self._data_provider = _SingleFrameDataProvider(layouts_lazy)
        elif isinstance(layouts_lazy, pl.DataFrame):
            self._data_provider = _SingleFrameDataProvider(layouts_lazy.lazy())
        elif isinstance(layouts_lazy, Iterable):
            self._data_provider = _MultiFrameDataProvider(layouts_lazy)
        else:
            raise ValueError("Must be lazy frame or iterable of lazy frames")

        if not partitioning:
            partitioning = PreComputedLayouts.DEFAULT_PARTITIONING
        self._partitioning = partitioning

    @staticmethod
    def create_empty():
        """Create an empty PreComputedLayouts instance."""
        return PreComputedLayouts(None)

    def __repr__(self) -> str:
        """Return a string representation of the PreComputedLayouts instance."""
        has_data = "empty" if self.is_empty else "with data"
        return f"PreComputedLayouts({has_data})"

    def __str__(self) -> str:
        """Return a string representation of the PreComputedLayouts instance."""
        return self.__repr__()

    @property
    def is_empty(self) -> bool:
        """Return True if the layout is empty."""
        return self._data_provider.is_empty()

    @property
    def partitioning(self) -> list[str]:
        """Return the partitioning of the layouts.

        This is used to determine how to partition the data when
        serializing it to disk, for the cases when this is applicable.
        For example when writing hive style parquet files.
        """
        return self._partitioning

    def write_parquet(self, path: Path, partitioning: list[str]) -> None:
        """Write a parquet file to the provided path."""
        self._data_provider.write_parquet(path, partitioning)

    def unique_components(self) -> set[str]:
        """Return the unique components in the layouts."""
        return self._data_provider.unique_components()

    def to_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        """Return the layouts as a pandas DataFrame.

        :param columns: the columns to return, if `None` all columns will be returned
        :return: A pandas DataFrame with the layout(s)
        """
        return self._data_provider.to_df(columns)

    @property
    def lazy(self) -> pl.LazyFrame:
        """Return the layouts as a polars LazyFrame."""
        return self._data_provider.lazy()

    def filter(
        self,
        component_ids: str | set[str] | None = None,
        graph_projection: str | set[str] | None = None,
        layout_method: str | set[str] | None = None,
    ) -> PreComputedLayouts:
        """Filter the layouts based on the provided criteria.

        :param component_ids: the component ids to filter on
        :param graph_projection: the graph projection to filter on
        :param layout_method: the layout method to filter on
        :return: A new PreComputedLayouts instance with the filtered layouts
        :rtype: PreComputedLayouts
        """
        return PreComputedLayouts(
            self._data_provider.filter(component_ids, graph_projection, layout_method),
            partitioning=self.partitioning,
        )

    def component_iterator(
        self,
        component_ids: Optional[set[str] | str] = None,
        graph_projections: Optional[set[str] | str] = None,
        layout_methods: Optional[set[str] | str] = None,
        columns: Optional[list[str]] = None,
    ) -> Iterable[pd.DataFrame]:
        """Get an iterator over the components provided.

        If `component_ids` is not provided, all components will be returned.

        Providing additional parameters will filter the layouts based on these
        criteria.

        :param component_ids: the component ids to filter on, if `None` all components
                              will be returned.
        :param graph_projections: the graph projections to filter on
        :param layout_methods: the layout methods to filter on
        :param columns: the columns to return, if `None` all columns will be returned
        :yields pd.DataFrame: A generator over the components where each dataframe contains the layout(s)
                  for that component
        """
        if not component_ids:
            unique_components = self._data_provider.unique_components()
        else:
            unique_components = self._convert_to_set(component_ids)  # type: ignore

        # We read in batches since it makes the read operations slightly
        # faster than iterating them one at the time
        for component_ids in batched(unique_components, 20):
            data = self.filter(
                component_ids=component_ids,
                graph_projection=graph_projections,
                layout_method=layout_methods,
            )

            for _, df in data.lazy.collect().group_by("component"):
                yield df.select(columns if columns else pl.all()).to_pandas()

    @staticmethod
    def _convert_to_set(
        value: Optional[str | set[str]] = None,
    ) -> Optional[set[str]]:
        if value is not None:
            value = set([value]) if isinstance(value, str) else set(value)
        return value

    def copy(self) -> PreComputedLayouts:
        """Copy the PreComputedLayouts instance."""
        return PreComputedLayouts(
            self.lazy.clone(), partitioning=self.partitioning.copy()
        )


def aggregate_precomputed_layouts(
    pxl_datasets: Iterable[tuple[str, PixelDataset | None]],
    all_markers: set[str],
) -> PreComputedLayouts:
    """Aggregate precomputed layouts into a single PreComputedLayouts instance."""

    def zero_fill_missing_markers(
        lazyframe: pl.LazyFrame, all_markers: set[str]
    ) -> pl.LazyFrame:
        missing_markers = all_markers - set(lazyframe.collect_schema().names())
        return lazyframe.with_columns(
            **{marker: pl.lit(0) for marker in missing_markers}
        )

    def data():
        for sample_name, pxl_dataset in pxl_datasets:
            layout = pxl_dataset.precomputed_layouts
            if layout is None:
                continue
            if layout.is_empty:
                continue
            layout_with_name = layout.lazy.with_columns(
                sample=pl.lit(sample_name),
                component=pl.concat_str(
                    pl.col("component"), pl.lit(sample_name), separator="_"
                ),
            ).pipe(zero_fill_missing_markers, all_markers=all_markers)
            yield layout_with_name

    try:
        return PreComputedLayouts(
            data(),
            partitioning=["sample"] + PreComputedLayouts.DEFAULT_PARTITIONING,
        )
    except ValueError:
        return PreComputedLayouts.create_empty()


def _zero_fill_missing_markers(dataframe: pd.DataFrame, all_markers) -> pd.DataFrame:
    missing_markers = all_markers - set(dataframe.columns.to_list())
    for marker in missing_markers:
        dataframe[marker] = 0
    return dataframe


def _compute_layouts(
    edgelist: pl.DataFrame,
    add_node_marker_counts: bool,
    layout_algorithms: list[SupportedLayoutAlgorithm] | SupportedLayoutAlgorithm,
    all_markers: list[str],
):
    if isinstance(layout_algorithms, str):
        layout_algorithms = [layout_algorithms]

    def data():
        for component, df in edgelist.group_by("component"):
            logger.debug("Computing for component %s", component)
            graph = Graph.from_edgelist(
                df.lazy(),
                add_marker_counts=add_node_marker_counts,
                simplify=True,
                use_full_bipartite=True,
                convert_indices_to_integers=False,
            )
            for layout_algorithm in layout_algorithms:
                yield (
                    pl.DataFrame(
                        graph.layout_coordinates(
                            layout_algorithm=layout_algorithm,
                            get_node_marker_matrix=add_node_marker_counts,
                            cache=False,
                            only_keep_a_pixels=False,
                        )
                        .pipe(
                            _zero_fill_missing_markers
                            # If we don't want to add the marker counts
                            # we just pass this through an identity function
                            if add_node_marker_counts
                            else lambda df, all_markers: df,
                            all_markers=all_markers,
                        )
                        .sort_index(axis=1)
                    )
                    .with_columns(
                        component=pl.lit(component[0]),
                        graph_projection=pl.lit("bipartite"),
                        layout=pl.lit(layout_algorithm),
                    )
                    .lazy()
                )

    result = pl.concat(data(), how="diagonal")
    return result


def _wrap_get_layouts(d):
    """Deconstruct a tuple of args to their corresponding argument.

    This is only necessary since imap does not support multiple arguments
    """
    (edgelist, add_node_marker_counts, layout_algorithms, all_markers) = d
    return _compute_layouts(
        edgelist, add_node_marker_counts, layout_algorithms, all_markers
    )


def _data(
    pixel_dataset: PixelDataset,
    components,
    add_node_marker_counts,
    layout_algorithms,
    all_markers,
):
    with get_pool_executor() as executor:
        # Batching over the component groups and setting a higher chunksize
        # to speeds up the computation here. I have not extensively
        # benchmarked the exact values here, but they seem to work well
        # enough
        component_group = map(set, batched(components, 20))
        yield from executor.imap(
            _wrap_get_layouts,
            (
                (
                    pixel_dataset.edgelist_lazy.filter(
                        pl.col("component").is_in(components)
                    ).collect(),
                    add_node_marker_counts,
                    layout_algorithms,
                    all_markers,
                )
                for components in component_group
            ),
            chunksize=1,
        )


@experimental
def generate_precomputed_layouts_for_components(
    pixel_dataset: PixelDataset,
    components: set[str] | None = None,
    add_node_marker_counts: bool = True,
    layout_algorithms: SupportedLayoutAlgorithm
    | list[SupportedLayoutAlgorithm] = "wpmds_3d",
) -> PreComputedLayouts:
    """Generate precomputed layouts for the components in the PixelDataset.

    :param pixel_dataset: the PixelDataset to generate the layouts for
    :param components: the components to generate the layouts for, if `None` all components will be used
    :param add_node_marker_counts: whether to add the marker counts to the layout. If you don't need them
                                   you can set this to false and speed up the computations.
    :param layout_algorithm: the layout algorithm to use
    """
    if components is None:
        components = set(pixel_dataset.adata.obs.index)

    # This is a bit of a hack to handle that different
    # graphs might have different markers, but for the end result
    # to be valid we need to have the same markers across all
    # components
    all_markers = set(pixel_dataset.adata.var.index.to_list())

    return PreComputedLayouts(
        _data(
            pixel_dataset=pixel_dataset,
            components=components,
            add_node_marker_counts=add_node_marker_counts,
            layout_algorithms=layout_algorithms,
            all_markers=all_markers,
        )
    )



"""Routines for printing a report.

Copyright © 2024 Pixelgen Technologies AB.
"""

from typing import Any, Dict

import pydantic
from cutadapt.report import Statistics

from pixelator.pna.read_processing.statistics import HasCustomStatistics
from pixelator.pna.report.models import SampleReport


class BarcodeCorrectionStatistics(Statistics):
    """Collect statistics for the barcode correction step."""

    FILTERS = {"invalid_antibody_barcode": "no valid antibody barcodes found"}

    def __init__(self) -> None:
        """Initialize the BarcodeCorrectionStatistics collector."""
        super().__init__()
        self.paired = False
        self._custom_stats: dict[str, Any] = dict()

    def __iadd__(self, other: Any):
        """Merge statistics from another object into this one."""
        super().__iadd__(other)
        if hasattr(other, "_custom_stats"):
            for name, value in other._custom_stats.items():
                if name in self._custom_stats:
                    self._custom_stats[name] += value
                else:
                    self._custom_stats[name] = value

        return self

    @property
    def pid_pair_counter(self):
        """Return the number of reads per PID pair."""
        return self._custom_stats["demux"].get_pid_group_counter()

    def as_json(self, gc_content: float = 0.5, one_line: bool = False) -> Dict:
        """Return a dict representation suitable for dumping in JSON format.

        To achieve a more compact representation, set one_line to True, which
        will wrap some items in a `cutadapt.json.OneLine` object, and use
        `cutadapt.json.dumps` instead of `json.dumps` to dump the dict.
        """
        demux_report = None

        assert "demux" in self._custom_stats
        demux_report = self._custom_stats["demux"].collect()
        return demux_report

    def _collect_step(self, step):
        super()._collect_step(step)
        if isinstance(step, HasCustomStatistics):
            name = step.get_statistics_name()
            if name in self._custom_stats:
                self._custom_stats[name] += step.get_statistics()
            else:
                self._custom_stats[name] = step.get_statistics()

    def collect(
        self,
        n: int,
        total_bp1: int,
        total_bp2: int | None,
        modifiers,
        steps,
        set_paired_to_none: bool = False,
    ):
        """Enable stats.paired to be set to None when unknown."""
        stats = super().collect(n, total_bp1, total_bp2, modifiers, steps)
        if set_paired_to_none:
            stats.paired = None
        return stats


class DemuxSampleReport(SampleReport):
    """Model for a demux sample report."""

    input_reads: int = pydantic.Field(
        ..., description="The number of input reads processed."
    )
    output_reads: int = pydantic.Field(
        ...,
        description="The number of reads that have valid marker barcodes and are demuxed.\nThis corresponds to passed_exact_reads + passed_corrected_reads.",
    )
    failed_reads: int = pydantic.Field(
        ...,
        description="The number of reads with barcodes that could not be corrected against the panel.",
    )
    output_exact_reads: int = pydantic.Field(
        ...,
        description="The number of reads for which the antibody barcodes were an exact match with the panel.",
    )
    output_corrected_reads: int = pydantic.Field(
        ...,
        description="The number of reads for which the antibody barcodes needed error correction before matching with the panel.",
    )
    invalid_pid1_reads: int = pydantic.Field(
        ...,
        description="The number of amplicon reads that where discarded because of an unrecognised PID1 region.",
    )
    invalid_pid2_reads: int = pydantic.Field(
        ...,
        description="The number of amplicon reads that where discarded because of an unrecognised PID2 region.",
    )
    invalid_pid1_pid2_reads: int = pydantic.Field(
        ...,
        description="The number of amplicon reads that where discarded because of both the PID1 and PID2 regions being unrecognised.",
    )
    pid1_matches_distance_distribution: dict[int, int] = pydantic.Field(
        ..., description="The distribution of distances for PID1 matches."
    )
    pid2_matches_distance_distribution: dict[int, int] = pydantic.Field(
        ..., description="The distribution of distances for PID2 matches."
    )
    pid_group_sizes: list[tuple[str, str, int]] = pydantic.Field(
        ..., description="The number of reads per PID pair."
    )

    @pydantic.computed_field(  # type: ignore
        description="The fraction of reads that were discarded by the demux step.",
        return_type=float,
    )
    @property
    def fraction_failed_reads(self) -> float:
        """Calculate the fraction of reads that were discarded by the demux step."""
        if self.input_reads == 0:
            return 0
        return self.failed_reads / self.input_reads



"""Module for computing localization proximity statistics.

Copyright © 2024 Pixelgen Technologies AB
"""

from typing import Callable, List, Literal

import numpy as np
import pandas as pd
import polars as pl
from scipy.stats import mannwhitneyu, norm
from statsmodels.stats.multitest import multipletests

from pixelator.pna.analysis.permute import edgelist_permutations


def get_join_counts(edgelist: pl.DataFrame) -> pd.DataFrame:
    """Compute the number of edges for each marker pair in the given edgelist.

    Args:
        edgelist (pl.DataFrame): A DataFrame representing the edgelist with
            columns "marker_1" and "marker_2".

    Returns:
        pd.DataFrame: A DataFrame containing the number of edges for each
        marker pair. The resulting DataFrame includes the columns:
            - "marker_1": The first marker in the pair.
            - "marker_2": The second marker in the pair.
            - "join_count": The number of edges between the marker pair.

    """
    pair_cnt = edgelist.group_by(["marker_1", "marker_2"]).len().to_pandas()
    m1 = pair_cnt["marker_1"].astype(str)
    m2 = pair_cnt["marker_2"].astype(str)
    pair_cnt.loc[m1 > m2, "marker_1"] = m2
    pair_cnt.loc[m1 > m2, "marker_2"] = m1
    all_markers = list(set(pair_cnt["marker_1"]).union(set(pair_cnt["marker_2"])))
    pair_cnt["marker_1"] = pd.Categorical(pair_cnt["marker_1"], categories=all_markers)
    pair_cnt["marker_2"] = pd.Categorical(pair_cnt["marker_2"], categories=all_markers)
    pair_cnt = (
        pair_cnt.groupby(["marker_1", "marker_2"], observed=False)["len"]
        .sum()
        .reset_index()
    )
    pair_cnt.rename(columns={"len": "join_count"}, inplace=True)
    pair_cnt = pair_cnt[
        pair_cnt["marker_1"].astype(str) <= pair_cnt["marker_2"].astype(str)
    ]  # Each marker-pair gets only one row, where marker_1/marker_2 are in lexographic order
    return pair_cnt


def _get_markers_above_min_count(edgelist: pl.DataFrame, min_count: int = 0) -> set:
    """Filter out markers with low counts from the edgelist.

    Args:
        edgelist (pl.DataFrame): A DataFrame representing the edgelist.
        min_count (int, optional): Minimum count threshold for markers. Defaults to 0.

    Returns:
        pl.DataFrame: A filtered DataFrame with low-count markers removed.

    """
    umi1_counts = (
        edgelist.select(["umi1", "marker_1"])
        .unique()
        .group_by("marker_1")
        .len()
        .rename({"marker_1": "marker", "len": "umi1_count"})
    )
    umi2_counts = (
        edgelist.select(["umi2", "marker_2"])
        .unique()
        .group_by("marker_2")
        .len()
        .rename({"marker_2": "marker", "len": "umi2_count"})
    )
    umi_counts = (
        umi1_counts.join(umi2_counts, on="marker", how="full", coalesce=True)
        .fill_null(0)
        .with_columns(total_count=pl.col("umi1_count") + pl.col("umi2_count"))
    )

    passing_markers = umi_counts.filter(pl.col("total_count") >= min_count)

    return set(passing_markers["marker"])


def proximity_with_permute_stats(
    edgelist: pl.DataFrame,
    proximity_function: Callable[[pl.DataFrame], pd.DataFrame],
    result_columns: list[str],
    n_permutations: int = 100,
    seed: int | None = 42,
    min_std: float = 1.0,
    min_marker_count: int = 0,
) -> pd.DataFrame:
    """Compute proximity results augmented with statistics based on permutation tests.

    This function calculates proximity metrics for a given edgelist and augments
    the results with statistical measures derived from permutation tests. It
    supports computing z-scores and p-values for specified result columns.

    Args:
        edgelist (pl.DataFrame): A DataFrame representing the edgelist.
        proximity_function (Callable[[pl.DataFrame], pd.DataFrame]): A function
            that computes proximity metrics for the given edgelist.
        result_columns (list[str]): A list of column names for which statistics
            (e.g., z-scores, p-values) will be computed.
        n_permutations (int, optional): The number of permutations to perform.
            Defaults to 100.
        seed (int | None, optional): Seed for the random number generator.
            Defaults to 42.
        min_std (float, optional): Minimum standard deviation to use when
            normalizing z-scores. Defaults to 1.0.
        min_marker_count (int, optional): Minimum marker count threshold for
            filtering the edgelist. Defaults to 0.

    Returns:
        pd.DataFrame: A DataFrame containing the proximity results augmented with
        statistical measures, including expected means, standard deviations,
        z-scores, and p-values for the specified result columns.

    """
    passing_markers = _get_markers_above_min_count(edgelist, min_marker_count)
    results = proximity_function(edgelist).set_index(["marker_1", "marker_2"])

    def compute_permuted_results():
        permutations = edgelist_permutations(edgelist, n_permutations, seed)
        for idx, perm in enumerate(permutations):
            perm_results = proximity_function(perm)
            perm_results["perm_idx"] = idx
            yield perm_results

    permuted_results = pd.concat(compute_permuted_results())

    for col in result_columns:
        results[f"{col}_expected_mean"] = permuted_results.groupby(
            ["marker_1", "marker_2"], observed=False
        )[col].mean()
        results[f"{col}_expected_sd"] = permuted_results.groupby(
            ["marker_1", "marker_2"], observed=False
        )[col].std()
        results[f"{col}_expected_mean"] = results[f"{col}_expected_mean"].fillna(0)
        results[f"{col}_expected_sd"] = results[f"{col}_expected_sd"].fillna(min_std)
        results[f"{col}_z"] = (
            results[col] - results[f"{col}_expected_mean"]
        ) / np.maximum(results[f"{col}_expected_sd"], min_std)
        results[f"{col}_p"] = norm.sf(np.abs(results[f"{col}_z"]))

    results = results.reset_index()
    results = results[
        results["marker_1"].isin(passing_markers)
        & results["marker_2"].isin(passing_markers)
    ]

    return results


def jcs_with_permute_stats(
    edgelist: pl.DataFrame,
    n_permutations: int = 100,
    min_marker_count: int = 0,
) -> pd.DataFrame:
    """Compute proximity results augmented with statistics based on permutation tests.

    Args:
        edgelist (pl.DataFrame): A DataFrame representing the edgelist.
        n_permutations (int, optional): Number of permutations to perform. Defaults to 100.
        min_marker_count (int, optional): Minimum marker count to consider. Defaults to 0.

    Returns:
        pd.DataFrame: A DataFrame containing the proximity statistics.

    """
    return proximity_with_permute_stats(
        edgelist,
        get_join_counts,
        ["join_count"],
        n_permutations=n_permutations,
        seed=42,
        min_std=1.0,
        min_marker_count=min_marker_count,
    )


def _filter_target_data(
    proximity_df, contrast_column, reference, target, metric, min_n_obs
):
    target_data = proximity_df[proximity_df[contrast_column].isin([reference, target])]

    if min_n_obs > 0:
        group_counts = (
            target_data.groupby(["marker_1", "marker_2", contrast_column])
            .size()
            .unstack(fill_value=0)
        )
        valid_markers = group_counts[
            (group_counts[reference] > min_n_obs) & (group_counts[target] > min_n_obs)
        ].index
        target_data = target_data[
            target_data.set_index(["marker_1", "marker_2"]).index.isin(valid_markers)
        ]
    return target_data


def _perform_mannwhitneyu_test(ref_group, tgt_group):
    u_stat, p_value = mannwhitneyu(ref_group, tgt_group, alternative="two-sided")
    auc = u_stat / (len(ref_group) * len(tgt_group))
    tgt_median = np.median(tgt_group)
    ref_median = np.median(ref_group)
    median_diff = tgt_median - ref_median
    return u_stat, p_value, auc, median_diff, tgt_median, ref_median


def calculate_differential_proximity(
    proximity_df: pd.DataFrame,
    contrast_column: str,
    reference: str,
    targets: List[str] | None = None,
    metric: str = "join_count_z",
    metric_type: Literal["all", "self", "co"] = "all",
    min_n_obs: int = 0,
    p_adjust_method: Literal[
        "bonferroni", "holm", "hochberg", "hommel", "fdr_bh", "fdr_by", "sidak"
    ] = "bonferroni",
) -> pd.DataFrame:
    """Perform differential analysis on marker-pair proximity data.

    Args:
        proximity_df (pd.DataFrame): Input data containing proximity metrics and
            grouping information. Must include columns for `contrast_column`,
            `marker_1`, `marker_2`, and the proximity metric (default: "join_count_z").
        contrast_column (str): The column name representing the grouping variable
        reference (str): The reference group in the `contrast_column`.
        targets (List[str] | None, optional): List of target groups to compare
            against the reference. If None, all groups in `contrast_column` except
            the reference are used. Defaults to None.
        metric (str, optional): Column name representing the proximity metric to
            analyze. Defaults to "join_count_z".
        metric_type (Literal["all", "self", "co"], optional): Type of measures to
            analyze ("self", "co", or "all" proximities). Defaults to "all".
        min_n_obs (int, optional): Minimum number of observations required for a
            group to be included in the analysis. Defaults to 0.
        p_adjust_method (optional): Method for adjusting p-values
            for multiple comparisons. Defaults to "bonferroni".

    Returns:
        pd.DataFrame: A DataFrame containing the results of the differential
        proximity analysis, including statistical metrics and adjusted p-values.

    Raises:
        ValueError: If `contrast_column` is not in `proximity_df`.
        ValueError: If no data is found for the specified `metric_type`.

    """
    if contrast_column not in proximity_df.columns:
        raise ValueError(f"{contrast_column} must be a column in the data.")

    if targets is None:
        targets = proximity_df[contrast_column].unique().tolist()
        targets.remove(reference)

    if metric_type == "self":
        proximity_df = proximity_df[
            proximity_df["marker_1"] == proximity_df["marker_2"]
        ]
    elif metric_type == "co":
        proximity_df = proximity_df[
            proximity_df["marker_1"] != proximity_df["marker_2"]
        ]

    if proximity_df.empty:
        raise ValueError("No data found for the specified metric type.")

    def calc_targets_differential():
        for target in targets:
            target_data = _filter_target_data(
                proximity_df, contrast_column, reference, target, metric, min_n_obs
            )

            for (marker_1, marker_2), group in target_data.groupby(
                ["marker_1", "marker_2"]
            ):
                ref_group = group[group[contrast_column] == reference][metric]
                tgt_group = group[group[contrast_column] == target][metric]

                if len(ref_group) == 0 or len(tgt_group) == 0:
                    continue

                u_stat, p_value, auc, median_diff, tgt_median, ref_median = (
                    _perform_mannwhitneyu_test(ref_group, tgt_group)
                )
                results = pd.Series(
                    {
                        "marker_1": marker_1,
                        "marker_2": marker_2,
                        "reference": reference,
                        "target": target,
                        "u_stat": u_stat,
                        "p_value": p_value,
                        "auc": auc,
                        "median_diff": median_diff,
                        "tgt_median": tgt_median,
                        "ref_median": ref_median,
                        "n_ref": len(ref_group),
                        "n_tgt": len(tgt_group),
                    }
                )
                yield results

    results_df = pd.DataFrame((calc_targets_differential()))
    if results_df.empty:
        return results_df
    results_df["p_adjusted"] = multipletests(
        results_df["p_value"], method=p_adjust_method
    )[1]

    return results_df



"""Module with functions for created permuted data.

Copyright © 2023 Pixelgen Technologies AB.
"""

import time
from typing import Generator, Optional

import pandas as pd
from numpy.random import Generator as RandomNumberGenerator
from numpy.random import default_rng

from pixelator.mpx.analysis.types import (
    RegionByCountsDataFrame,
)


def _get_random_number_generator(
    random_seed: Optional[int] = None,
) -> RandomNumberGenerator:
    if not random_seed:
        random_seed = int(time.time())
    return default_rng(seed=random_seed)


def permutations(
    df: RegionByCountsDataFrame, n: int = 50, random_seed: Optional[int] = None
) -> Generator[RegionByCountsDataFrame, None, None]:
    """Generate `n` permutations of the data provided in `df`.

    :param df: dataframe to use as basis of permutations
    :param n: number of permutations to generate, defaults to 50
    :param random_seed: set a seed to the random number generator
                        needed to make results deterministic, defaults to None
    :yields RegionByCountsDataFrame: a RegionByCountsDataFrame object
    :return: a generator that yields `n` permutations of the input dataframe
    :rtype: Generator[RegionByCountsDataFrame, None, None]
    """
    random_number_generator = _get_random_number_generator(random_seed)
    for _ in range(n):
        yield permute(df, random_number_generator)


def permute(
    df: RegionByCountsDataFrame,
    random_number_generator: Optional[RandomNumberGenerator] = None,
) -> RegionByCountsDataFrame:
    """Generate permutations of data frame.

    Permute the given dataframe in a way that preserves the number of
    counts in each region. The proportions of each marker is kept
    approximately by sampling from a multinomial distribution of the
    original counts. Note that especially for low count markers this
    might mean that some markers that have counts in the original
    dataframe might not have it counts in the permuted dataframe.

    :param df: input dataframe to permute
    :param random_number_generator: Set a random number generator
                                    to make results reproducible, defaults to None
    :return: a permuted dataframe
    """
    if not random_number_generator:
        random_number_generator = _get_random_number_generator()

    marker_counts = df.sum(axis="index")
    marker_sum = marker_counts.sum()
    marker_probs = marker_counts / marker_sum
    marker_probs_sum = marker_probs.sum()
    assert -1e-10 < 1 - marker_probs_sum and marker_probs_sum - 1 < 1e-10, (
        f"{marker_probs_sum}"
    )
    region_counts = df.sum(axis="columns")

    sampling = random_number_generator.multinomial(
        n=region_counts.to_numpy().reshape((len(region_counts), 1)), pvals=marker_probs
    ).reshape(df.shape)

    return pd.DataFrame(
        sampling,
        columns=df.columns.copy(),
        index=df.index.copy(),
    )



"""Protocol of graph backends used by pixelator.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Protocol,
    Set,
    Tuple,
    Union,
)

import pandas as pd
import polars as pl
from scipy.sparse import csr_matrix

if TYPE_CHECKING:
    from pixelator.mpx.graph import Graph

SupportedLayoutAlgorithm = Literal[
    "fruchterman_reingold",
    "fruchterman_reingold_3d",
    "kamada_kawai",
    "kamada_kawai_3d",
    "pmds",
    "pmds_3d",
    "wpmds_3d",
]


class GraphBackend(Protocol):
    """Protocol for graph backends."""

    @staticmethod
    def from_edgelist(
        edgelist: Union[pd.DataFrame, pl.LazyFrame],
        add_marker_counts: bool,
        simplify: bool,
        use_full_bipartite: bool,
        convert_indices_to_integers: bool = True,
    ) -> GraphBackend:
        """Build a graph from an edgelist.

        Build a Graph from an edge list (pd.DataFrame). Multiple options are available
        to build the graph, `add_marker_counts` will add a dictionary of marker counts
        to each node, `simplify` will remove redundant edges and `use_full_bipartite`
        will not project the graph (UPIA).

        The graph will contain the edge attributes present in the edge list when
        `use_full_bipartite` is True and a dictionary of marker counts in each
        vertex (node) when `add_marker_counts` is True. If `use_full_bipartite` is
        False or `simplify` is True the edge attributes will be lost.

        :param edgelist: the edge list (dataframe) corresponding to the graph
        :param add_marker_counts: add a dictionary of marker counts to each node
        :param simplify: simplifies the graph (remove redundant edges)
        :param use_full_bipartite: use the bipartite graph instead of the projection
                                  (UPIA)
        :param convert_indices_to_integers: convert the indices to integers (this is the default)
        :returns: a Graph instance
        :rtype: GraphBackend
        :raises: AssertionError when the input edge list is not valid
        """
        ...

    @staticmethod
    def from_raw(graph: Any) -> GraphBackend:
        """Generate a Graph from a valid Graph object.

        What the valid graph object is depends on the underlying
        GraphBackend implementation.

        Typically what you want to use is `from_edgelist`.

        :param graph: input graph to use
        :return: A pixelator GraphBackend object
        :rtype: GraphBackend
        """
        ...

    @property
    def raw(self):
        """Get the raw underlying graph representation."""
        ...

    @property
    def vs(self) -> VertexSequence:
        """Get a sequence of the vertices in the Graph instance."""
        ...

    @property
    def es(self) -> EdgeSequence:
        """A sequence of the edges in the Graph instance."""
        ...

    def vcount(self) -> int:
        """Get the total number of vertices in the Graph instance."""
        ...

    def ecount(self) -> int:
        """Get the total number of edges in the Graph instance."""
        ...

    def get_adjacency_sparse(
        self, node_ordering: Iterable[Any] | None = None
    ) -> csr_matrix:
        """Get the sparse adjacency matrix.

        :param node_ordering: Control the node ordering in the adjacency matrix
        :return: a sparse adjacency matrix
        """
        ...

    def connected_components(self) -> VertexClustering:
        """Get the connected components in the Graph instance."""
        ...

    def node_marker_counts(self) -> pd.DataFrame:
        """Get the marker counts of each node as a dataframe.

        :return: node markers as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if graph nodes don't include markers
        """
        ...

    def layout_coordinates(
        self,
        layout_algorithm: SupportedLayoutAlgorithm = "wpmds_3d",
        only_keep_a_pixels: bool = True,
        get_node_marker_matrix: bool = True,
        random_seed: Optional[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Generate coordinates and (optionally) node marker counts for plotting.

        Generate a dataframe with coordinates, and (optionally) node marker
        counts to use that can be used for plotting.

        The layout options are:
          - fruchterman_reingold
          - fruchterman_reingold_3d
          - kamada_kawai
          - kamada_kawai_3d
          - pmds
          - pmds_3d
          - wpmds_3d

        For most cases the `pmds` options should be about 10-100x faster
        than the force directed layout methods, i.e. `fruchterman_reingold`
        and `kamada_kawai`. Among the force directed layout methods,
        `fruchterman_reingold` is generally faster than `kamada_kawai`. The
        `wpmds_3d` method uses edge weights to improve the layout, but is slightly
        slower than `pmds_3d`.

        :param layout_algorithm: the layout algorithm to use to generate the coordinates
        :param only_keep_a_pixels: If true, only keep the a-pixels
        :param get_node_marker_matrix: Add a matrix of marker counts to each
                                       node if True.
        :param random_seed: used as the seed for graph layouts with a stochastic
                            element. Useful to get deterministic layouts across
                            method calls.
        :param **kwargs: will be passed to the underlying layout implementation
        :return: the coordinates and markers (if activated) as a dataframe
        :rtype: pd.DataFrame
        :raises: AssertionError if the provided `layout_algorithm` is not valid
        :raises: ValueError if the provided current graph instance is empty
        """
        ...

    def get_edge_dataframe(self) -> pd.DataFrame:
        """Get the edges as a pandas DataFrame."""
        ...

    def get_vertex_dataframe(self) -> pd.DataFrame:
        """Get all vertices as a pandas DataFrame."""
        ...

    def add_edges(self, edges: Iterable[Tuple[int]]) -> None:
        """Add edges to the graph instance.

        :param edges: Add the following edges to the graph instance.
        """
        ...

    def add_vertices(self, n_vertices: int, attrs: Dict[str, List]) -> None:
        """Add some number of vertices to the graph instance.

        :param n_vertices: the number of vertices to be added to the graph instance.
        :param attrs: dict of sequences, all of length equal to the number of vertices
                      to be added, containing the attributes of the new vertices. If
                      `n_vertices=1` then they have to be lists of length 1.
        :raises IndexError: if the number of graph vertices to add and lists of
                            attributes are of different lengths
        """
        ...

    def add_names_to_vertexes(self, vs_names: List[str]) -> None:
        """Rename the current vertices on the graph instance.

        :param vs_names: Add the following vertices to the graph instance.
        :raises ValueError: if the graph is empty
        :raises IndexError: if the number of graph vertices and list of names are
                            of different length
        """
        ...


class Vertex(Protocol):
    """Protocol for a single vertex in a graph."""

    @property
    def index(self) -> int:
        """Get the index of the vertex."""
        ...

    def __getitem__(self, attr: str) -> Any:
        """Get the value of the vertex attribute `attr`."""
        ...

    def __str__(self) -> str:
        """Get a string representation of the vertex."""
        return f"Vertex({self.index})"

    def __repr__(self) -> str:
        """Get a representation of the vertex."""
        return str(self)


class Edge(Protocol):
    """Protocol for edges in a graph."""

    @property
    def vertex_tuple(self) -> Tuple[Vertex, Vertex]:
        """Return the vertices the edge connects as a tuple."""
        ...

    @property
    def index(self) -> int:
        """The index of the edge."""
        ...

    def __str__(self) -> str:
        """Get a string representation of the edge."""
        return f"Edge({self.vertex_tuple})"

    def __repr__(self) -> str:
        """Get a representation of the edge."""
        return str(self)


class VertexSequence(Iterable[Vertex], Protocol):
    """A sequence of vertexes from a graph."""

    def vertices(self) -> Iterable[Vertex]:
        """Get an iterable of vertices."""
        ...

    def __len__(self) -> int:
        """Get the number of vertexes."""
        ...

    def __iter__(self) -> Iterator[Vertex]:
        """Get an iterator over the vertices in the sequence."""
        ...

    def attributes(self) -> Set[str]:
        """Get all attributes associated with the vertices."""
        ...

    def get_vertex(self, vertex_id: int) -> Vertex:
        """Get the Vertex corresponding to the vertex id."""
        ...

    def get_attribute(self, attr: str) -> Iterable[Any]:
        """Get the values of the attribute."""
        ...


class EdgeSequence(Iterable[Edge], Protocol):
    """A sequence of edges from a graph."""

    def __len__(self) -> int:
        """Get the number of edges."""
        ...

    def __iter__(self) -> Iterator[Edge]:
        """Get an iterator over the edges."""
        ...

    def select_where(self, key, value) -> EdgeSequence:
        """Select a subset of edges where `key == value`."""
        ...

    def select_within(self, component):
        """Select all edges that are part of `component`."""
        ...


class VertexClustering(Iterable[VertexSequence], Protocol):
    """A cluster of vertexes, such as a community in a graph."""

    def __len__(self) -> int:
        """Get the number of clusters."""
        ...

    def __iter__(self) -> Iterator[VertexSequence]:
        """Provide an iterator over the clusters."""
        ...

    @property
    def modularity(self) -> float:
        """Get the modularity of the clusters."""
        ...

    def crossing(self) -> EdgeSequence:
        """Get any crossing edges."""
        ...

    def giant(self) -> Graph:
        """Get the largest component."""
        ...

    def subgraphs(self) -> Iterable[Graph]:
        """Get subgraphs of each cluster."""
        ...



"""Copyright © 2023 Pixelgen Technologies AB."""

from .cli_info import CLIInvocationInfo, CommandInfo, CommandOption
from .reporting import PixelatorReporting
from .workdir import PixelatorWorkdir, SingleCellStage, WorkdirOutputNotFound

__all__ = [
    "CLIInvocationInfo",
    "CommandInfo",
    "CommandOption",
    "PixelatorWorkdir",
    "PixelatorReporting",
    "SingleCellStage",
    "WorkdirOutputNotFound",
]



"""Connected components module.

Functions and classes relating to building connected components in the graph step.

Copyright © 2024 Pixelgen Technologies AB
"""

import logging
import typing
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable, Literal

import networkx as nx
import numpy as np
import pandas as pd
import polars as pl
import xxhash
from graspologic_native import leiden

from pixelator.common.annotate.aggregates import call_aggregates
from pixelator.common.annotate.cell_calling import find_component_size_limits
from pixelator.common.exceptions import PixelatorBaseException
from pixelator.pna.anndata import pna_edgelist_to_anndata
from pixelator.pna.config import PNAAntibodyPanel
from pixelator.pna.graph.report import (
    GraphStatistics,
)
from pixelator.pna.pixeldataset import PNAPixelDataset
from pixelator.pna.pixeldataset.io import PixelFileWriter

LEIDEN_RANDOM_SEED = 1
MIN_PNA_COMPONENT_SIZE = 8000

logger = logging.getLogger(__name__)


class ConnectedComponentException(PixelatorBaseException):
    """Exception raised when there is an issue with connected components."""


def _calculate_component_stats(
    connected_components: Iterable[set[int]],
) -> tuple[int, float]:
    if not isinstance(connected_components, list):
        connected_components = list(connected_components)
    max_component_size = 0
    total_graph_size = 0
    nbr_of_components = 0
    for component in connected_components:
        nbr_of_components += 1
        total_graph_size += len(component)
        max_component_size = max(max_component_size, len(component))
    return (
        nbr_of_components,
        max_component_size / total_graph_size if total_graph_size else 0,
    )


def _filter_edgelist(
    edgelist: pl.LazyFrame,
    min_read_count: int,
    component_stats: GraphStatistics,
):
    """Filter the edgelist by read count."""
    edgelist = edgelist.filter(pl.col("read_count") >= min_read_count)
    post_filter_stat = edgelist.select(
        [
            pl.col("read_count").sum(),
            pl.col("uei_count").sum(),
            pl.col("umi1").n_unique().alias("umi1_count"),
            pl.col("umi2").n_unique().alias("umi2_count"),
            pl.len().alias("edge_count"),
        ]
    ).collect()
    component_stats.molecules_post_read_count_filtering = post_filter_stat["uei_count"][
        0
    ]
    component_stats.reads_post_read_count_filtering = post_filter_stat["read_count"][0]
    component_stats.node_count_pre_recovery = (
        post_filter_stat["umi1_count"] + post_filter_stat["umi2_count"]
    )[0]
    component_stats.edge_count_pre_recovery = post_filter_stat["edge_count"][0]
    return edgelist


def _label_connected_components(
    edgelist: pl.LazyFrame,
    component_stats: GraphStatistics,
):
    """Build the graph and calculate some statistics."""
    logger.debug("Finding connected components")

    graph = nx.from_edgelist(edgelist.select(["umi1", "umi2"]).collect().rows())
    connected_components = list(nx.connected_components(graph))
    umi_component_map = dict()
    for i, cc in enumerate(connected_components):
        umi_component_map.update(dict.fromkeys(cc, i))
    del graph

    (
        component_count,
        fraction_nodes_in_largest_component,
    ) = _calculate_component_stats(connected_components)
    del connected_components

    if component_count == 0:
        raise ConnectedComponentException(
            "No connected components found in the graph. Will not continue with the rest of the computations."
        )

    component_stats.component_count_pre_recovery = component_count
    component_stats.fraction_nodes_in_largest_component_pre_recovery = (
        fraction_nodes_in_largest_component
    )

    return umi_component_map, component_stats


@dataclass
class RefinementOptions:
    """Options for refining components."""

    min_component_size: int = 10

    max_edges_to_remove: int | None = 20
    max_edges_to_remove_relative: float | None = None
    leiden_resolution: float = 1.0
    min_component_size_to_prune: int = 100


@dataclass
class StagedRefinementOptions:
    """Options for refining components at each stage."""

    inital_stage_options: RefinementOptions = field(
        default_factory=lambda: RefinementOptions(
            max_edges_to_remove=None,
            max_edges_to_remove_relative=None,
            leiden_resolution=1.0,
        )
    )
    refinement_stage_options: RefinementOptions = field(
        default_factory=lambda: RefinementOptions(
            max_edges_to_remove=4,
            max_edges_to_remove_relative=None,
            leiden_resolution=0.01,
        )
    )
    max_component_refinement_depth: int = 1


@dataclass(slots=True)
class MultipletRecoveryStats:
    """Statistics from multiplet recovery."""

    crossing_edges_removed: int = 0
    crossing_edges_removed_in_initial_stage: int = 0
    max_recursion_depth: int = 0


def merge_communities_with_many_crossing_edges(
    edgelist: pl.DataFrame,
    node_community_dict: dict,
    max_edges_to_remove: int | None = None,
    max_edges_to_remove_relative: float | None = None,
) -> pd.Series:
    """Merge communities what have many crossing edges between them in an edge list.

    This function takes an edge list and a dictionary with the community mapping
    for each node. It then computes the number of edges between communities and
    if they are higher than the given a threshold. It assigns the same community
    id to the nodes in the connected strongly connected communities. The
    threshold is determined by an absolute count and/or relative to the number
    of nodes in the smaller of the two communities, whichever is higher.
    If one of them is None, only the other one is considered and if they are
    both None, the split communities are not considered for merging.

    :param edgelist: The edge list to process
    :param node_community_dict: A dictionary mapping each node to a community
    :param n_edges: The threshold for the number of edges to be found between
    communities to merge or None to avoid merging
    :returns: The updated community mapping
    """
    community_serie = pd.Series(node_community_dict)
    if max_edges_to_remove is None and max_edges_to_remove_relative is None:
        return community_serie
    if community_serie.nunique() == 1:
        return community_serie

    community_sizes = community_serie.value_counts().to_dict()
    # This will be an edgelist where every node is a community
    # and there is an edge when there are more than maximum_edges to
    # remove between the communities.
    community_edgelist = (
        edgelist.with_columns(
            community1=pl.col("umi1").replace_strict(node_community_dict),
            community2=pl.col("umi2").replace_strict(node_community_dict),
        )
        .with_columns(
            community1=pl.min_horizontal(pl.col("community1"), pl.col("community2")),
            community2=pl.max_horizontal(pl.col("community1"), pl.col("community2")),
        )
        .filter(pl.col("community1") != pl.col("community2"))
        .group_by(["community1", "community2"])
        .len()
        .with_columns(
            community1_size=pl.col("community1").replace_strict(community_sizes),
            community2_size=pl.col("community2").replace_strict(community_sizes),
        )
        .with_columns(
            threshold=pl.max_horizontal(
                max_edges_to_remove,
                max_edges_to_remove_relative
                * pl.min_horizontal(
                    pl.col("community1_size"), pl.col("community2_size")
                ),
            ),
        )
        .filter(pl.col("len") >= pl.col("threshold"))
        .select(["community1", "community2"])
        .rows()
    )

    communities_graph = nx.from_edgelist(community_edgelist)
    for cc in nx.connected_components(communities_graph):
        new_tag = min(cc)
        community_serie[community_serie.isin(cc)] = new_tag

    return community_serie


def _get_umi_component_map_from_edgelist(edgelist: pl.LazyFrame) -> dict:
    """Get a umi component map from an edgelist."""
    umi_component_map = dict()
    if "component1" in edgelist.collect_schema().names():
        component_umis = (
            edgelist.group_by("component1")
            .agg(pl.col("umi1").unique(), pl.col("umi2").unique())
            .collect()
        )
        for comp, umi1, umi2 in component_umis.rows():
            umi_component_map.update(dict.fromkeys(umi1 + umi2, comp))

    else:
        umis = set(edgelist.select("umi1").unique().collect()["umi1"]).union(
            set(edgelist.select("umi2").unique().collect()["umi2"])
        )
        umi_component_map.update(dict.fromkeys(umis, 0))
    return umi_component_map


def _update_components_column(
    edgelist: pl.LazyFrame, umi_component_map: dict
) -> pl.LazyFrame:
    """Update the component column in an edgelist."""
    return edgelist.with_columns(
        component1=pl.col("umi1").replace_strict(umi_component_map),
        component2=pl.col("umi2").replace_strict(umi_component_map),
    )


def make_edgelist_with_component_column(
    edgelist: pl.LazyFrame, umi_component_map: dict
) -> pl.LazyFrame:
    """Add a component column to an edgelist using a node map and remove crossing edges.

    :param edgelist: The edgelist to add the component column to.
    :param umi_component_map: A dictionary mapping nodes to components.
    :returns: The edgelist with crossing edges removed and the component column
    added.
    """
    return (
        _update_components_column(edgelist, umi_component_map)
        .filter(pl.col("component1") == pl.col("component2"))
        .drop("component2")
        .rename({"component1": "component"})
    )


def _find_initial_components_to_be_refined(
    edgelist: pl.LazyFrame, min_component_size: int
) -> list[str]:
    return list(
        edgelist.group_by("component1")
        .agg(
            pl.col("umi1").n_unique().alias("n_umi1"),
            pl.col("umi2").n_unique().alias("n_umi2"),
        )
        .filter((pl.col("n_umi1") + pl.col("n_umi2")) > min_component_size)
        .select("component1")
        .unique()
        .collect()["component1"]
    )


def _refine_components(
    umi_component_map: dict,
    clusters_to_be_refined: pl.dataframe.group_by.GroupBy,
    leiden_iterations: int,
    refinement_options: RefinementOptions,
) -> tuple[dict, list]:
    def id_generator(start=0):
        next_id = start
        while True:
            yield next_id
            next_id += 1

    id_gen = id_generator(max(umi_component_map.values()) + 1)

    refinement_candidates = []
    for _, cluster_edges in clusters_to_be_refined:
        number_of_nodes = cluster_edges.n_unique("umi1") + cluster_edges.n_unique(
            "umi2"
        )
        if number_of_nodes < refinement_options.min_component_size:
            continue

        _, leiden_communities = leiden(
            cluster_edges.select(
                pl.col("umi1").cast(pl.Utf8), pl.col("umi2").cast(pl.Utf8)
            )
            .with_columns(weight=pl.lit(1))
            .rows(),
            seed=LEIDEN_RANDOM_SEED,
            use_modularity=True,
            resolution=refinement_options.leiden_resolution,
            # These parameters are used to sync up the native implementation with
            # the python implementation we originally used.
            iterations=leiden_iterations + 1,
            randomness=0.001,
            trials=1,
            starting_communities=None,
        )
        # Map the communites back from strings to integers
        leiden_communities = {int(k): v for k, v in leiden_communities.items()}  # type: ignore
        community_serie = merge_communities_with_many_crossing_edges(
            cluster_edges,
            leiden_communities,
            refinement_options.max_edges_to_remove,
            refinement_options.max_edges_to_remove_relative,
        )
        if community_serie.nunique() > 1:
            for _, nodes in community_serie.groupby(community_serie):
                community_id = next(id_gen)
                umi_component_map.update(dict.fromkeys(nodes.index, community_id))
                if len(nodes) > refinement_options.min_component_size:
                    refinement_candidates.append(community_id)
    return umi_component_map, refinement_candidates


def recover_multiplets(
    edgelist: pl.LazyFrame,
    umi_component_map: dict | None = None,
    leiden_iterations: int = 1,
    refinement_options: StagedRefinementOptions | None = None,
) -> tuple[nx.Graph, MultipletRecoveryStats]:
    """Recovery multiplets by leiden community detection, removing crossing edges between communities."""
    if umi_component_map is None:
        umi_component_map = _get_umi_component_map_from_edgelist(edgelist)

    if refinement_options is None:
        refinement_options = StagedRefinementOptions(
            inital_stage_options=RefinementOptions(),
            refinement_stage_options=RefinementOptions(),
        )

    edgelist_with_components = _update_components_column(edgelist, umi_component_map)
    to_be_refined = _find_initial_components_to_be_refined(
        edgelist_with_components,
        refinement_options.inital_stage_options.min_component_size,
    )

    max_recursion_depth_observed = 0
    for recursion_level in range(refinement_options.max_component_refinement_depth):
        logger.debug(
            "Running Leiden community detection. At recursion depth %s", recursion_level
        )

        clusters_to_be_refined = (
            edgelist_with_components.filter(pl.col("component1").is_in(to_be_refined))
            # Remove the crossing edges, will do nothing on the first iteration
            .filter(pl.col("component1") == pl.col("component2"))
            .collect()
            .group_by("component1")
        )
        umi_component_map, to_be_refined = _refine_components(
            umi_component_map=umi_component_map,  # type: ignore
            clusters_to_be_refined=clusters_to_be_refined,
            leiden_iterations=leiden_iterations,
            refinement_options=(
                refinement_options.inital_stage_options
                if recursion_level == 0
                else refinement_options.refinement_stage_options
            ),
        )
        edgelist_with_components = _update_components_column(
            edgelist,
            umi_component_map,  # type: ignore
        )
        if recursion_level == 0:
            nbr_of_edges_removed_in_initial_stage = (
                edgelist_with_components.filter(
                    pl.col("component1") != pl.col("component2")
                )
                .select(pl.len())
                .collect()
            )["len"][0]

        if not to_be_refined:
            logger.debug(
                "No more components to refine, will break the refinement loop."
            )
            max_recursion_depth_observed = recursion_level
            break

    total_nbr_crossing_edges = (
        edgelist_with_components.filter(pl.col("component1") != pl.col("component2"))
        .select(pl.len())
        .collect()
    )["len"][0]
    return umi_component_map, MultipletRecoveryStats(
        crossing_edges_removed=total_nbr_crossing_edges,
        crossing_edges_removed_in_initial_stage=nbr_of_edges_removed_in_initial_stage,
        max_recursion_depth=max_recursion_depth_observed,
    )


def filter_components_by_size_dynamic(
    component_sizes: pd.DataFrame,
    lowest_passable_bound: int | None = MIN_PNA_COMPONENT_SIZE,
) -> tuple[pd.DataFrame, int | None]:
    """Filter components by size using dynamic thresholds.

    :param component_sizes: A DataFrame with columns `component` and `n_umi`.
    :returns: only the 'component' column of with the components that pass the filter, and the lower bound.
    """
    if lowest_passable_bound is None:
        lowest_passable_bound = MIN_PNA_COMPONENT_SIZE

    lower_bound = find_component_size_limits(
        component_sizes=component_sizes["n_umi"].to_numpy(), direction="lower"
    )
    if lower_bound is None or lower_bound < lowest_passable_bound:
        lower_bound = lowest_passable_bound
        logger.warning(
            "Could not find a lower bound for component size filtering, will "
            "set the lower bound to " + str(lowest_passable_bound),
        )
    return (
        component_sizes.filter(pl.col("n_umi") >= lower_bound)["component"],
        lower_bound,
    )


def filter_components_by_size_hard_thresholds(
    component_sizes: pd.DataFrame, lower_bound: int | None, higher_bound: int | None
) -> pd.DataFrame:
    """Filter components by size using hard thresholds.

    :param component_sizes: A DataFrame with columns `component` and `n_umi`.
    :param lower_bound: The lower bound for the component size.
    :param higher_bound: The higher bound for the component size.
    :returns: only the 'component' column of with the components that pass the filter.
    """
    if lower_bound is None:
        lower_bound = 0
    if higher_bound is None:
        higher_bound = np.iinfo(np.uint64).max
    return component_sizes.filter(
        (pl.col("n_umi") >= lower_bound) & (pl.col("n_umi") <= higher_bound)
    )["component"]


def hash_component(component: set[int]) -> str:
    """Hash a component. Should yield the same hash if the components consists of the same nodes."""
    hasher = xxhash.xxh3_64()
    for node in sorted(component):
        hasher.update(node.to_bytes(length=8, byteorder="little"))
    return hasher.hexdigest()


def _find_clashing_umis(
    molecules_lazy_frame: pl.LazyFrame,
) -> tuple[pl.Series, pl.Series]:
    # Find umis that are associated with multiple markers
    # indicating a clash.
    umi1_clashes = (
        molecules_lazy_frame.select(["umi1", "marker_1"])
        .unique()
        .group_by("umi1")
        .len()
        .filter(pl.col("len") > 1)
        .collect()["umi1"]
    )
    umi2_clashes = (
        molecules_lazy_frame.select(["umi2", "marker_2"])
        .unique()
        .group_by("umi2")
        .len()
        .filter(pl.col("len") > 1)
        .collect()["umi2"]
    )

    # Find umis that are present in both umi1 and umi2
    # which also indicates a clash
    umi1_umi2_clashes = (
        pl.concat(
            (
                molecules_lazy_frame.select("umi1").unique().rename({"umi1": "umi"}),
                molecules_lazy_frame.select("umi2").unique().rename({"umi2": "umi"}),
            )
        )
        .group_by("umi")
        .len()
        .filter(pl.col("len") > 1)
        .collect()["umi"]
    )
    umi1_clashes = umi1_clashes.extend(umi1_umi2_clashes)
    umi2_clashes = umi2_clashes.extend(umi1_umi2_clashes)
    return umi1_clashes, umi2_clashes


def build_pxl_file_with_components(
    molecules_lazy_frame: pl.LazyFrame,
    panel: PNAAntibodyPanel,
    sample_name: str,
    path_output_pxl_file: Path,
    multiplet_recovery: bool,
    leiden_iterations: int,
    min_count: int,
    refinement_options: StagedRefinementOptions | None = None,
    component_size_threshold: tuple[int, int] | bool = True,
) -> tuple[PNAPixelDataset, GraphStatistics]:
    """Create a pxl file after having created components and removed crossing edges."""
    with TemporaryDirectory(prefix="pixelator-") as tmp_dir:
        tmp_dir_path = Path(tmp_dir)

        edgelist_with_components, component_statics = find_components(  # type: ignore
            input_edgelist=molecules_lazy_frame,
            multiplet_recovery=multiplet_recovery,
            leiden_iterations=leiden_iterations,
            min_read_count=min_count,
            return_component_statistics=True,
            refinement_options=refinement_options,
            component_size_threshold=component_size_threshold,
        )
        logger.debug("Sorting edgelist by component")
        edgelist_with_components_sorted = edgelist_with_components.sort("component")
        tmp_edgelist_file_sorted = (
            tmp_dir_path / f"{sample_name}.edgelist.sorted.parquet"
        )
        edgelist_with_components_sorted.sink_parquet(tmp_edgelist_file_sorted)
        with PixelFileWriter(path_output_pxl_file) as pxl_file_writer:
            pxl_file_writer.write_edgelist(tmp_edgelist_file_sorted)
            logger.debug("Counting molecules")

            sums = (
                edgelist_with_components_sorted.select(["uei_count", "read_count"])
                .sum()
                .collect()
            )
            component_statics.molecules_output = sums.item(0, "uei_count")

            logger.debug("Building edgelist from anndata")
            adata = pna_edgelist_to_anndata(
                pxl_file_writer.get_connection(), panel=panel
            )
            call_aggregates(adata)

            component_statics.reads_output = adata.obs["reads_in_component"].sum()

            component_statics.median_reads_per_component = adata.obs[
                "reads_in_component"
            ].median()
            component_statics.median_markers_per_component = adata.obs["n_umi"].median()

            # Add tau_type metrics
            aggregates_mask = adata.obs["tau_type"] != "normal"
            number_of_aggregates = np.sum(aggregates_mask)

            component_statics.aggregate_count = number_of_aggregates
            aggregate_stats = (
                adata[aggregates_mask]
                .obs[["n_edges", "n_umi", "reads_in_component"]]
                .sum()
            )

            component_statics.read_count_in_aggregates = aggregate_stats[
                "reads_in_component"
            ].item()
            component_statics.node_count_in_aggregates = aggregate_stats["n_umi"].item()
            component_statics.edge_count_in_aggregates = aggregate_stats[
                "n_edges"
            ].item()

            # Sort adata on component names for stable output
            adata = adata[adata.obs_names.sort_values(), :]

            # import here to avoid circular imports
            from pixelator import __version__

            metadata = {
                "sample_name": sample_name,
                "version": __version__,
                "technology": "single-cell-pna",
                "panel_name": panel.name,
                "panel_version": panel.version,
            }

            logger.debug("Building pxl file")

            pxl_file_writer.write_metadata(metadata)
            pxl_file_writer.write_adata(adata)

        return PNAPixelDataset.from_pxl_files(path_output_pxl_file), component_statics


def _get_working_edgelist(
    input_edgelist: pl.LazyFrame,
) -> tuple[pl.LazyFrame, pl.DataFrame]:
    """Get a working edgelist and a map from original to working node names."""
    node_map = (
        pl.concat(
            (
                input_edgelist.select("umi1").rename({"umi1": "original_name"}),
                input_edgelist.select("umi2").rename({"umi2": "original_name"}),
            )
        )
        .unique()
        .sort("original_name")
        .with_row_index()
        .rename({"index": "working_name"})
        .collect()
    )

    working_edgelist = input_edgelist.with_columns(
        pl.col("umi1").replace_strict(
            old=node_map["original_name"], new=node_map["working_name"]
        ),
        pl.col("umi2").replace_strict(
            old=node_map["original_name"], new=node_map["working_name"]
        ),
    )
    return working_edgelist, node_map


def _add_post_recovery_stats(edgelist: pl.LazyFrame, component_stats: GraphStatistics):
    """Add post recovery statistics to the component statistics object."""
    post_recovery_stats = edgelist.select(
        [
            pl.col("umi1").n_unique().alias("umi1_count"),
            pl.col("umi2").n_unique().alias("umi2_count"),
            pl.len().alias("edge_count"),
        ]
    ).collect()
    component_stats.node_count_post_recovery = (
        post_recovery_stats["umi1_count"] + post_recovery_stats["umi2_count"]
    )[0]
    component_stats.edge_count_post_recovery = post_recovery_stats["edge_count"][0]

    connected_component_sizes = (
        edgelist.group_by("component")
        .agg(
            pl.col("umi1").n_unique().alias("n_umi1"),
            pl.col("umi2").n_unique().alias("n_umi2"),
        )
        .select(component_size=pl.col("n_umi1") + pl.col("n_umi2"))
        .collect()
    )
    component_stats.component_count_post_recovery = connected_component_sizes.shape[0]
    component_stats.fraction_nodes_in_largest_component_post_recovery = (
        connected_component_sizes["component_size"].max()  # type: ignore
        / connected_component_sizes["component_size"].sum()  # type: ignore
    )
    return component_stats


def _name_components_with_umi_hashes(edgelist):
    comp_umis = (
        edgelist.group_by("component")
        .agg(pl.col("umi1").unique(), pl.col("umi2").unique())
        .collect()
    )
    comp_hashes = dict()
    for comp, umi1, umi2 in comp_umis.rows():
        comp_hashes[comp] = hash_component(set(umi1 + umi2))

    return edgelist.with_columns(pl.col("component").replace_strict(comp_hashes))


def _remove_umi_clashes_and_get_stats(input_edgelist, component_stats):
    raw_stat = input_edgelist.select(["read_count", "uei_count"]).sum().collect()
    component_stats.molecules_input = raw_stat["uei_count"][0]
    component_stats.reads_input = raw_stat["read_count"][0]
    umi1_clashes, umi2_clashes = _find_clashing_umis(input_edgelist)
    no_clash_edgelist = input_edgelist.filter(
        ~pl.col("umi1").is_in(umi1_clashes) & ~pl.col("umi2").is_in(umi2_clashes)
    )
    post_collision_stat = (
        no_clash_edgelist.select(["read_count", "uei_count"]).sum().collect()
    )
    component_stats.molecules_post_umi_collision_removal = post_collision_stat[
        "uei_count"
    ][0]
    component_stats.reads_post_umi_collision_removal = post_collision_stat[
        "read_count"
    ][0]

    return no_clash_edgelist, component_stats


@typing.overload
def find_components(
    input_edgelist: pl.LazyFrame,
    multiplet_recovery: bool,
    leiden_iterations: int = 10,
    min_read_count: int = 2,
    component_size_threshold: tuple[int, int] | bool = True,
    refinement_options: StagedRefinementOptions | None = None,
    return_component_statistics: Literal[True] = True,
    dynamic_lowest_passable_bound: int | None = None,
) -> tuple[pl.LazyFrame, GraphStatistics]: ...


@typing.overload
def find_components(
    input_edgelist: pl.LazyFrame,
    multiplet_recovery: bool,
    leiden_iterations: int = 10,
    min_read_count: int = 2,
    component_size_threshold: tuple[int, int] | bool = True,
    refinement_options: StagedRefinementOptions | None = None,
    return_component_statistics: Literal[False] = False,
    dynamic_lowest_passable_bound: int | None = None,
) -> pl.LazyFrame: ...


def find_components(
    input_edgelist: pl.LazyFrame,
    multiplet_recovery: bool,
    leiden_iterations: int = 10,
    min_read_count: int = 2,
    component_size_threshold: tuple[int, int] | bool = True,
    refinement_options: StagedRefinementOptions | None = None,
    return_component_statistics: bool = False,
    dynamic_lowest_passable_bound: int | None = None,
) -> pl.LazyFrame | tuple[pl.LazyFrame, GraphStatistics]:
    """Retrieve all connected components from an edgelist.

    This function takes as input an edge list in `parquet` format that has
    been generated with `pixelator collapse`. The function filters the
    edge list by count (`min_count`) and then adds a `component` column to the
    edge list with the respective connected components ids obtained from the graph.

    If `multiplet_recovery` is True the edge list is then processed to recover
    components from technical multiplets. The recovery is done using community
    detection to detect and remove problematic edges using the Leiden [1]_ community
    detection algorithm.

    .. [1] Traag, V.A., Waltman, L. & van Eck, N.J. From Louvain to Leiden: guaranteeing
        well-connected communities. Sci Rep 9, 5233 (2019).
        https://doi.org/10.1038/s4q:598-019-41695-z

    :param input_edgelist: The input edgelist
    :param multiplet_recovery: if True run multiplet recovery, otherwise skip it
    :param leiden_iterations: number of Leiden iterations to run
    :param min_read_count: minimum number of supporting reads for an edge to be considered part of the graph
    :param component_size_threshold: if True, filter components by dynamic size thresholds, if a tuple, filter by hard thresholds as (min, max)
    :param refinement_options: options for component refinement
    :param return_component_statistics: if True, return a component statistics object

    :returns: an edgelist with components added to it, and a component statistics object if return_component_statistics is True
    """
    component_stats = GraphStatistics()
    no_clash_edgelist, component_stats = _remove_umi_clashes_and_get_stats(
        input_edgelist, component_stats
    )
    working_edgelist, node_map = _get_working_edgelist(no_clash_edgelist)
    working_edgelist = _filter_edgelist(
        working_edgelist, min_read_count, component_stats
    )
    logger.debug("Labelling connected components")
    umi_component_map, component_stats = _label_connected_components(
        working_edgelist, component_stats
    )

    if multiplet_recovery:
        logger.debug("Initiating multiplet recovery")
        umi_component_map, recover_stats = recover_multiplets(
            working_edgelist,
            umi_component_map,
            leiden_iterations,
            refinement_options=refinement_options,
        )
        component_stats.crossing_edges_removed = recover_stats.crossing_edges_removed
        component_stats.crossing_edges_removed_initial_stage = (
            recover_stats.crossing_edges_removed_in_initial_stage
        )
        component_stats.max_recursion_depth = recover_stats.max_recursion_depth

    edgelist_with_components = make_edgelist_with_component_column(
        working_edgelist, umi_component_map
    )
    if multiplet_recovery:
        component_stats = _add_post_recovery_stats(
            edgelist_with_components, component_stats
        )
        component_stats.component_count_pre_component_size_filtering = (
            component_stats.component_count_post_recovery
        )
    else:
        component_stats.component_count_pre_component_size_filtering = (
            component_stats.component_count_pre_recovery
        )

    # Restore original umi names
    edgelist_with_components = edgelist_with_components.with_columns(
        pl.col("umi1").replace_strict(
            old=node_map["working_name"], new=node_map["original_name"]
        ),
        pl.col("umi2").replace_strict(
            old=node_map["working_name"], new=node_map["original_name"]
        ),
    )

    edgelist_with_components = _filter_connected_components_by_size(
        edgelist_with_components,
        component_size_threshold,
        component_stats,
        dynamic_lowest_passable_bound,
    )
    edgelist_with_components = _name_components_with_umi_hashes(
        edgelist_with_components
    )
    if return_component_statistics:
        return edgelist_with_components, component_stats
    return edgelist_with_components


def _filter_connected_components_by_size(
    edgelist: pl.LazyFrame,
    component_size_threshold,
    component_stats: GraphStatistics,
    dynamic_lowest_passable_bound=None,
):
    """Filter connected components by size and get statistics."""
    component_sizes = (
        edgelist.group_by("component")
        .agg(
            pl.col("umi1").n_unique().alias("n_umi1"),
            pl.col("umi2").n_unique().alias("n_umi2"),
        )
        .select(pl.col("component"), n_umi=pl.col("n_umi1") + pl.col("n_umi2"))
        .collect()
    )
    unique, counts = np.unique(
        component_sizes["n_umi"].cast(pl.Int32), return_counts=True
    )
    component_stats.pre_filtering_component_sizes = dict(zip(unique, counts))
    if isinstance(component_size_threshold, bool) and component_size_threshold:
        logger.debug("Filtering components by dynamic size thresholds")
        try:
            passing_components, min_size = filter_components_by_size_dynamic(
                component_sizes,
                lowest_passable_bound=dynamic_lowest_passable_bound,
            )
            component_stats.component_size_min_filtering_threshold = min_size
        except Exception as e:
            # This is a hack since the exception is not properly
            # by passed the spline interpolation package.
            if str(type(e)) == "<class 'dfitpack.error'>":
                msg = (
                    "Could not find component size filters, this probably means that no components (i.e. cells) were formed. "
                    "If you are running on the command line you can try to use --component-size-max-threshold and "
                    " --component-size-min-threshold to set hard thresholds, but most likely the problem is with the input data."
                )
                raise ConnectedComponentException(msg)
            else:
                raise e
    else:
        if component_size_threshold is False:
            min_size = 0
            max_size = np.iinfo(np.uint64).max
        else:
            min_size, max_size = component_size_threshold
        component_stats.component_size_max_filtering_threshold = max_size
        component_stats.component_size_min_filtering_threshold = min_size
        logger.debug(
            "Filtering components by hard size thresholds, min: %s, max: %s",
            min_size,
            max_size,
        )
        passing_components = filter_components_by_size_hard_thresholds(
            component_sizes, min_size, max_size
        )

    if len(passing_components) == 0:
        msg = (
            "No connected components found in the graph. Likely they were all filtered away for being to small. "
            "This indicates some serious issue with the data. Will not continue with the rest of the computations."
        )
        raise ConnectedComponentException(msg)

    component_stats.component_count_post_component_size_filtering = len(
        passing_components
    )

    edgelist = edgelist.filter(pl.col("component").is_in(passing_components))

    return edgelist



"""Graph console script for pixelator.

Copyright © 2024 Pixelgen Technologies AB.
"""

import sys

import click
import polars as pl

from pixelator.common.utils import (
    create_output_stage_dir,
    get_sample_name,
    log_step_start,
    sanity_check_inputs,
    timer,
    write_parameters_file,
)
from pixelator.pna.cli.common import logger, output_option, panel_option
from pixelator.pna.config import load_antibody_panel, pna_config
from pixelator.pna.graph.connected_components import (
    ConnectedComponentException,
    RefinementOptions,
    StagedRefinementOptions,
    build_pxl_file_with_components,
)
from pixelator.pna.graph.report import GraphSampleReport


@click.command(
    "graph",
    short_help=("compute graph, components and metrics from an edge list"),
    options_metavar="<options>",
)
@click.argument(
    "parquet_file",
    nargs=1,
    required=True,
    type=click.Path(exists=True),
    metavar="parquet",
)
@click.option(
    "--multiplet-recovery",
    required=False,
    is_flag=True,
    default=False,
    type=click.BOOL,
    help=("Activate the multiplet recovery using leiden community detection"),
)
@click.option(
    "--leiden-iterations",
    default=1,
    required=False,
    type=click.IntRange(1, 100),
    show_default=True,
    help=(
        "Number of iterations for the leiden algorithm, high values will decrease "
        "the variance of the results but increase the runtime"
    ),
)
@click.option(
    "--initial-stage-leiden-resolution",
    default=1.0,
    required=False,
    type=click.FloatRange(0, None),
    show_default=True,
    help=(
        "The resolution parameter for the leiden algorithm at the initial stage. This should "
        "typically be set higher than the refinement stage resolution."
    ),
)
@click.option(
    "--refinement-stage-leiden-resolution",
    default=0.01,
    required=False,
    type=click.FloatRange(0, None),
    show_default=True,
    help=(
        "The resolution parameter for the leiden algorithm at the refinement stage. This should "
        "typically be set lower than the initial stage resolution."
    ),
)
@click.option(
    "--min-count",
    default=1,
    required=False,
    type=click.IntRange(1, 50),
    show_default=True,
    help=(
        "Discard edges with a read count below given value. Set to 1 to disable filtering."
    ),
)
@click.option(
    "--min-component-size-in-refinement",
    default=1000,
    required=False,
    type=click.IntRange(min=1),
    show_default=True,
    help=("The minimum component size to consider for refinement"),
)
@click.option(
    "--max-refinement-recursion-depth",
    default=5,
    required=False,
    type=click.IntRange(min=1, max=20),
    show_default=True,
    help=(
        "The maximum recursion depth for the refinement algorithm. Set to 1 to disable refinement."
    ),
)
@click.option(
    "--initial-stage-max-edges-to-remove",
    default=None,
    required=False,
    type=click.IntRange(min=1, max=None),
    show_default=True,
    help=(
        "The maximum number of edges to remove between components during the "
        "initial stage (iteration == 0) of multiplet recovery."
    ),
)
@click.option(
    "--refinement-stage-max-edges-to-remove",
    default=4,
    required=False,
    type=click.IntRange(min=1, max=None),
    show_default=True,
    help=(
        "The maximum number of edges to remove between components during the "
        "refinement stage (iteration > 0) of multiplet recovery."
    ),
)
@click.option(
    "--initial-stage-max-edges-to-remove-relative",
    default=None,
    required=False,
    type=click.FloatRange(min=1e-6, max=None),
    show_default=True,
    help=(
        "The maximum number of edges to remove between two components relative "
        "to the number of nodes in the smaller of the two when during the "
        "initial stage (iteration == 0) of multiplet recovery."
    ),
)
@click.option(
    "--refinement-stage-max-edges-to-remove-relative",
    default=None,
    required=False,
    type=click.FloatRange(min=1e-6, max=None),
    show_default=True,
    help=(
        "The maximum number of edges to remove between two components relative "
        "to the number of nodes in the smaller of the two when during the "
        "refinement stage (iteration > 0) of multiplet recovery."
    ),
)
@click.option(
    "--min-component-size-to-prune",
    default=100,
    required=False,
    type=click.IntRange(min=1, max=None),
    show_default=True,
    help=(
        "The minimum number of nodes in an potential new components in order for it to be pruned."
    ),
)
@click.option(
    "--component-size-max-threshold",
    default=None,
    required=False,
    type=click.IntRange(min=1, max=None),
    show_default=True,
    help=(
        "Components with more nodes than this will be filtered from the output data. "
        "This is typically not needed. Setting this will disable the automatic size filtering."
    ),
)
@click.option(
    "--component-size-min-threshold",
    default=None,
    required=False,
    type=click.IntRange(min=1, max=None),
    show_default=True,
    help=(
        "Components with fewer nodes than this will be filtered from the output data. "
        "This is typically not needed. Setting this will disable the automatic size filtering."
    ),
)
@panel_option
@output_option
@click.pass_context
@timer
def graph(
    ctx,
    parquet_file,
    multiplet_recovery,
    leiden_iterations,
    initial_stage_leiden_resolution,
    refinement_stage_leiden_resolution,
    min_count,
    min_component_size_in_refinement,
    max_refinement_recursion_depth,
    initial_stage_max_edges_to_remove,
    refinement_stage_max_edges_to_remove,
    initial_stage_max_edges_to_remove_relative,
    refinement_stage_max_edges_to_remove_relative,
    min_component_size_to_prune,
    component_size_max_threshold,
    component_size_min_threshold,
    panel,
    output,
):
    """Find connected components from the input molecules.

    The graph stage will attempt to identify connected components from the input molecules
    in two stages.

    When `--multiple-recovery` is active we will try to break up components that are likely
    not real cells. We do so in two stages.

    In the initial stage will attempt to break up the so called mega cluster. This is a loosely
    connected but very large component that is often present in the data.

    In the subsequent refinement stage we inspect the resulting connected components and try
    to find components that can be split further, by identifying well-connected communities
    that have a low number of crossing edges between them. We will do so recursively until
    no more reasonable splits are found, or the maximum recursion depth is reached.


    After the connected components have been identified we will create a pxl file that contains
    data for all of there putative cells.
    """
    # log input parameters
    input_files = [parquet_file]
    log_step_start(
        "graph",
        input_files=input_files,
        output=output,
        multiplet_recovery=multiplet_recovery,
        leiden_iterations=leiden_iterations,
        initial_stage_leiden_resolution=initial_stage_leiden_resolution,
        refinement_stage_leiden_resolution=refinement_stage_leiden_resolution,
        min_count=min_count,
        min_component_size_in_refinement=min_component_size_in_refinement,
        max_refinement_recursion_depth=max_refinement_recursion_depth,
        initial_stage_max_edges_to_remove=initial_stage_max_edges_to_remove,
        refinement_stage_max_edges_to_remove=refinement_stage_max_edges_to_remove,
        initial_stage_max_edges_to_remove_relative=initial_stage_max_edges_to_remove_relative,
        refinement_stage_max_edges_to_remove_relative=refinement_stage_max_edges_to_remove_relative,
        min_component_size_to_prune=min_component_size_to_prune,
        component_size_max_threshold=component_size_max_threshold,
        component_size_min_threshold=component_size_min_threshold,
        panel=panel,
    )

    # some basic sanity check on the input files
    sanity_check_inputs(input_files, allowed_extensions="parquet")

    # create output folder if it does not exist
    graph_output = create_output_stage_dir(output, "graph")

    logger.info(f"Trying to identify components from file {parquet_file}")

    sample_name = get_sample_name(parquet_file)

    write_parameters_file(
        ctx,
        graph_output / f"{sample_name}.meta.json",
        command_path="pixelator single-cell-pna graph",
    )
    output_path = graph_output / f"{sample_name}.graph.pxl"

    panel = load_antibody_panel(pna_config, panel)
    initial_stage_refinement_options = RefinementOptions(
        min_component_size=min_component_size_in_refinement,
        max_edges_to_remove=initial_stage_max_edges_to_remove,
        max_edges_to_remove_relative=initial_stage_max_edges_to_remove_relative,
        min_component_size_to_prune=min_component_size_to_prune,
    )
    refinement_stage_refinement_options = RefinementOptions(
        min_component_size=min_component_size_in_refinement,
        max_edges_to_remove=refinement_stage_max_edges_to_remove,
        max_edges_to_remove_relative=refinement_stage_max_edges_to_remove_relative,
        min_component_size_to_prune=min_component_size_to_prune,
    )
    refinement_options = StagedRefinementOptions(
        max_component_refinement_depth=max_refinement_recursion_depth,
        inital_stage_options=initial_stage_refinement_options,
        refinement_stage_options=refinement_stage_refinement_options,
    )

    component_size_threshold = (
        (
            component_size_min_threshold,
            component_size_max_threshold,
        )
        if any([component_size_min_threshold, component_size_max_threshold])
        else True
    )

    if isinstance(component_size_threshold, tuple):
        logger.warning(
            "You have explicitly set component size thresholds. This is normally not needed. "
            "Please make sure you know what you are doing."
        )

    try:
        _, component_statics = build_pxl_file_with_components(
            molecules_lazy_frame=pl.scan_parquet(
                parquet_file, low_memory=True, cache=False
            ),
            panel=panel,
            sample_name=sample_name,
            path_output_pxl_file=output_path,
            multiplet_recovery=multiplet_recovery,
            leiden_iterations=leiden_iterations,
            min_count=min_count,
            refinement_options=refinement_options,
            component_size_threshold=component_size_threshold,
        )
    except ConnectedComponentException as e:
        logger.error(e)
        sys.exit(1)

    # build_pxl_file(edgelist_with_components, adata)
    metrics_file = graph_output / f"{sample_name}.report.json"
    report = GraphSampleReport(
        sample_id=sample_name,
        product_id="single-cell-pna",
        **component_statics.to_dict(),
    )
    report.write_json_file(metrics_file, indent=4)



"""Plugin for computing post analysis metrics for R&D report.

Copyright © 2024 Pixelgen Technologies AB.
"""

import logging
from copy import deepcopy
from typing import Literal

import networkx as nx
import numpy as np
import pandas as pd
import scipy as sp

from pixelator.common.exceptions import PixelatorBaseException
from pixelator.pna.analysis_engine import PerComponentTask
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset import PNAPixelDataset
from pixelator.pna.pixeldataset.io import PixelFileWriter, PxlFile

logger = logging.getLogger(__name__)


class SvdAnalysisError(PixelatorBaseException):
    """An exception for errors in SvdAnalysis.

    Currently used when trying to run SVD analysis on a graph with
    fewer nodes than the number of pivot points.
    """

    pass


DeadendOptions = Literal["remove", "ignore", "self-loop"]


def _mat_pow(mat, power):
    mat_power = mat
    for _ in range(power - 1):
        mat_power = mat @ mat_power
    return mat_power


def non_backtracking_transition_probabilities(
    graph: nx.Graph,
    n_steps: int = 3,
    cumulative: bool = True,
    deadend_action: DeadendOptions = "ignore",
    track_only_nonzero: bool = True,
) -> pd.Series:
    """Calculate the non-backtracking transition probabilities for an n_steps walk.

    A non-backtracking random walker can only transition on edges that are not the same as the
    edge it came from. The non-backtracking transition probabilities can be
    useful for identifying spurious edges in a graph. For example, a non-zero probability for
    a 3-step non-backtracking walk between two nodes i and j (connected by an edge) indicates
    that there is at least one path of length 3 connecting i and j. This can be an indication that
    i and j are located in the same neighborhood. If the probability is zero, there is no path of
    length 3 between i and j which is an indication that i and j are located in different neighborhoods.
    In the latter case, the edge between i and j is potentially spurious.

    One limitation with this method is that edges located in sparse regions of the graph often
    have zero probabilities. For example, when transitioning from i to j, if the end node j has
    a degree of 1, there are no paths of length > 1 and the probability will be zero for a k>1
    step non-backtracking walk. However, this is not an indication that i and j are located in
    different neighborhoods.

    :param graph: a networkx graph object.
    :param n_steps: the number of steps to take for the random walk.
    :param cumulative: if True, calculate the cumulative transition probabilities for all steps up to n_steps.
    :param deadend_action: the action to take for deadend nodes. Options are: "remove", "ignore", "self-loop".
    :param track_only_nonzero: if True, only track edges with non-zero probabilities without keeping the actual probablity.
    :return: a pandas series containing the transition probabilities.
    :raises ValueError: if the deadend_action is not one of the valid options.
    :raises ValueError: if the number of steps is less than 3.
    """
    valid_deadend_actions = ["remove", "ignore", "self-loop"]
    if deadend_action not in valid_deadend_actions:
        raise ValueError(
            f"Invalid value for deadend_action. Expected one of: {valid_deadend_actions}"
        )

    if n_steps < 3:
        raise ValueError("The number of steps must be at least 3.")

    if deadend_action == "remove":
        core_vals = nx.core_number(graph)
        to_remove = [i for i in core_vals.keys() if core_vals[i] == 1]
        graph = deepcopy(graph)
        graph.remove_nodes_from(to_remove)

    dir_g = nx.to_directed(graph)
    line_g = nx.line_graph(dir_g)

    # Remove edges that go back
    edgelist = list(line_g.edges)
    for edge in edgelist:
        if edge[0][0] == edge[1][1]:
            line_g.remove_edges_from([edge])

    adj_mat = nx.adjacency_matrix(line_g).astype(bool)

    if deadend_action == "self-loop":
        is_dead_end = (adj_mat.sum(axis=1) == 0).astype(float)
        adj_mat = adj_mat.tolil()
        adj_mat.setdiag(is_dead_end)
        adj_mat = adj_mat.tocsr().astype(bool)

    if track_only_nonzero:
        weighted_adj_mat = adj_mat
    else:
        weighted_adj_mat = (
            adj_mat / np.maximum(adj_mat.sum(axis=1).astype(float), 1)[:, None]
        )

    if cumulative:
        walk_probability = _mat_pow(weighted_adj_mat, 3)
        for i in range(3, n_steps):
            walk_probability += _mat_pow(weighted_adj_mat, i + 1)
    else:
        walk_probability = _mat_pow(weighted_adj_mat, n_steps)
    transition_probability = pd.Series(
        (adj_mat @ walk_probability).diagonal(), index=list(line_g.nodes)
    )

    return transition_probability


def svd_pivot_distances(
    g: nx.Graph, pivots: int = 100, seed: int | None = None
) -> tuple:
    """Single Value Decomposition of a graphs shortest path matrix.

    This function can be used to evaluate how well a graph can be represented
    in 3D space.

    Component graphs with high spatial resolution, i.e. where edges represent short distances
    in actual space, should be easier to embed in a euclidean space using multidimensional
    scaling techniques s.a. pMDS. In contrast, graphs with high spatial distortion or with
    a high amount of noise will be harder to embed. An example of such noise is the presence
    of edges which form long-range connections. Such connections will break the spatial
    structure of the graph and form shortest paths which are not representative of the actual
    spatial distances.

    The idea behind this method is that if a graph truly represents a spatial 3D structure,
    most of the variance should be explained by the first 3 singular vectors in the SVD. In
    the best case scenario, the first 3 singular vectors should explain 100% of the variance
    and all three singular vectors should contribute equally to the variance explained.

    The algorithm is outlined below:

    1. Select a number of pivot nodes at random.
    2. Calculate the shortest path length from the pivots to all other nodes.
    3. Double center the distance matrix.
    4. Perform SVD on the centered distance matrix.

    :param g: a networkx graph object
    :param pivots: the number of pivot nodes to use
    :param seed: the random seed to use
    :return: a tuple containing the results of the SVD
    :raises ValueError: the number of pivots must be less than the number of nodes in the graph
    """
    if pivots >= len(g.nodes):
        raise SvdAnalysisError(
            "'pivots' must be less than the number of nodes in the graph."
        )

    random_generator = np.random.default_rng(seed)

    node_list = list(g.nodes)

    # Select random pivot nodes
    pivs = random_generator.choice(node_list, pivots, replace=False)

    # Calculate the shortest path length from the pivots to all other nodes
    A = nx.to_scipy_sparse_array(g, weight=None, nodelist=node_list, format="csr")

    # Create shortest path matrix
    A.indices = A.indices.astype(np.intc, copy=False)
    A.indptr = A.indptr.astype(np.intc, copy=False)
    D = sp.sparse.csgraph.shortest_path(
        A,
        directed=False,
        unweighted=True,
        method="D",
        indices=np.where(np.isin(g.nodes, pivs))[0],
    ).T

    # Center values in rows and columns
    D2 = D**2
    cmean = np.mean(D2, axis=0)
    rmean = np.mean(D2, axis=1)
    D_pivs_centered = D2 - np.add.outer(rmean, cmean) + np.mean(D2)

    # Perform SVD on the centered distance matrix
    svd_res = np.linalg.svd(D_pivs_centered, full_matrices=False)

    return svd_res


def summarize_k_cores(g: nx.Graph) -> pd.DataFrame:
    """Summarize k-cores for a given component.

    Computes the coreness for all nodes in a given graph and summarizes the number
    of nodes per k-core in a table. The columns in the DataFrame are names 'k1', 'k2',
    ..., 'kM' where M is the maximum k-core value in the component. The k-core counts
    are useful for quantifying graph connectivity.

    :param g: a networkx Graph object
    :raises AssertionError: the input arguments are incorrect
    :return: a dataframe with the summarized node k-cores
    :rtype: pd.DataFrame
    """
    if not isinstance(g, nx.Graph):
        raise AssertionError("g must be a networkx graph object")

    k_core_nodes = pd.DataFrame.from_dict(
        nx.core_number(g), orient="index", columns=["k_core"]
    )

    average_k_core = k_core_nodes["k_core"].mean()

    # Summarize k-cores
    k_core_component_table = k_core_nodes.groupby("k_core").size()
    k_core_component_table = k_core_component_table.reset_index()
    k_core_component_table.columns = ["k_core", "count"]
    k_core_component_table["k_core"] = "k_core_" + k_core_component_table[
        "k_core"
    ].astype(str)
    k_core_component_table_wide = pd.pivot_table(
        k_core_component_table, values="count", columns="k_core"
    )
    k_core_component_table_wide.columns.name = None
    k_core_component_table_wide["average_k_core"] = average_k_core

    return k_core_component_table_wide


class KcoreAnalysis(PerComponentTask):
    """Run k-core decomposition on each component."""

    TASK_NAME = "k-core"

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> pd.DataFrame:
        """Run k-core analysis on a single component.

        Calls summarize_k_cores on a component graph and returns a pandas DataFrame with k-core counts.
        The columns in the DataFrame are names 'k_core_1', 'k_core_2', ..., 'k_core_M' where M is the
        maximum k-core value in the component. The k-core counts are useful for quantifying graph connectivity.

        :param component: a networkx graph for a component to run the analysis on.
        :param component_id: the id of the component.
        :return: a pandas DataFrame containing k-core counts.
        """
        logger.debug(f"Running k-core analysis on component {component_id}")
        k_core_summary = summarize_k_cores(
            g=component.raw,
        )
        k_core_summary.index = [component_id]
        return k_core_summary

    def add_to_pixel_file(self, data: pd.DataFrame, pxl_file_target: PxlFile) -> None:
        """Add k-core count data for all components to adata.obs in pxl_dataset.

        :param data: a pandas DataFrame containing k-core counts for all components.
        :param pxl_file_target: the PxlFile to add the data to.
        """
        logger.debug("Adding k-core analysis data to PNAPixelDataset")
        data.fillna(0, inplace=True)
        data.sort_index(axis="columns", inplace=True)

        adata = PNAPixelDataset.from_files(pxl_file_target).adata()
        adata.obs.drop(
            columns=[col for col in adata.obs.columns if "k_core_" in col],
            inplace=True,
            errors="ignore",
        )
        adata.obs.drop(columns=["average_k_core"], inplace=True, errors="ignore")
        adata.obs = adata.obs.join(data)
        with PixelFileWriter(pxl_file_target.path) as writer:
            writer.write_adata(adata)


class SvdAnalysis(PerComponentTask):
    """Run SVD variance explained on each component."""

    TASK_NAME = "SVD"

    def __init__(self, pivots: int = 100):
        """Initialize SvdAnalysis.

        :param pivots: the number of pivot points to use for SVD analysis.
        """
        self.pivots = pivots

    def run_on_component_graph(
        self, component: PNAGraph, component_id: str
    ) -> pd.DataFrame:
        """Run SVD analysis on component.

        Calls svd_pivot_distances on a component graph and returns a pandas DataFrame with variance
        explained for the first three singular vectors. The columns in the DataFrame are named 'svd_var_expl_s1',
        'svd_var_expl_s2', 'svd_var_expl_s3'. The variance explained is useful for investigating spatial coherence
        in the graph. In the best case scenario, the first three singular vectors contribute equally to the variance
        explained, and together they should explain 100% of the total variance explained.

        :param component: a networkx graph for a component to run the analysis on.
        :param component_id: the id of the component.
        :return: a pandas DataFrame containing variance explained for the first three singular vectors.
        If the SVD computation fails, e.g. if the number of pivot points is >= the number of nodes,
        the function returns a DataFrame with nan values.
        """
        logger.debug(f"Running SVD analysis on component {component_id}")

        try:
            # Compute SVD and fetch singular values
            _, s, _ = svd_pivot_distances(component.raw, seed=123, pivots=self.pivots)

            # Compute variance explained from singular values
            s_var = (s**2) / sum(s**2)

            svd_variance_explained = pd.DataFrame(
                {
                    "svd_var_expl_s1": [s_var[0]],
                    "svd_var_expl_s2": [s_var[1]],
                    "svd_var_expl_s3": [s_var[2]],
                },
                index=[component_id],
            )
        except SvdAnalysisError as e:
            # If SVD fails, return a DataFrame with nan values
            logger.warning(
                f"Failed to run SVD analysis on component {component_id}: {e}"
            )
            svd_variance_explained = pd.DataFrame(
                {
                    "svd_var_expl_s1": [np.nan],
                    "svd_var_expl_s2": [np.nan],
                    "svd_var_expl_s3": [np.nan],
                },
                index=[component_id],
            )

        return svd_variance_explained

    def add_to_pixel_file(self, data: pd.DataFrame, pxl_file_target: PxlFile) -> None:
        """Add svd variance explained for the first three singular vectors.

        Add svd variance explained for the first three singular vectors across
        all components to adata.obs in pxl_dataset.

        :param data: a pandas DataFrame containing svd variance explained for the first three
         singular vectors across all components.
        :param pxl_dataset: the PNAPixelDataset to add the data to.
        """
        logger.debug("Adding SVD analysis data to PNAPixelDataset")
        adata = PNAPixelDataset.from_files(pxl_file_target.path).adata()
        adata.obs.drop(
            columns=[col for col in adata.obs.columns if "svd_var_expl_" in col],
            inplace=True,
        )
        adata.obs = adata.obs.join(data)
        with PixelFileWriter(pxl_file_target.path) as writer:
            writer.write_adata(adata)
